from .base_attention_pooling import BaseAttentionPooling
from keras import saving


@saving.register_keras_serializable(package="Kerex.layers.pooling", name="AttentionPooling1D")
class AttentionPooling1D(BaseAttentionPooling):
    """
    1D attention based global pooling
    Each query represents a kind of feature or pattern in the input,
    which the input is attended to.

    Parameters
    ----------
    num_queries : int,
        Number of learned queries.
    num_heads : int, optional
        Number of heads in the multihead attention layer.
        Defaults to `4`.
    flatten : bool, optional
        If `True`, the output is flattened over the number of input channels and the number of queries.
        Defaults to `True`.
    kernel_initializer : str | keras.initializer.Initializer, optional
        Initializer for real- and imaginary weights.
        Defaults to `"orthogonal"`.
    kernel_regularizer : str | keras.regularizers.Regularizer, optional
        Kernel regularizer.
        Defaults to `None`.
    kernel_constraint : str | keras.constraints.Constraint, optional
        Kernel constraint.
        Defaults to `None`.
    name : str, optional
        Name of the model.
        If `None`, `name` is automatically inherited from the class name `"BaseSpectralConv"`.
        Defaults to `None`.

    Notes
    -----
    If `flatten=True`, this layer is basically a drop-in replacement for any global pooling layer,
    The output channels depend on the input channels and the number of queries:
    `out_channels = in_channels * num_queries`.

    """
    
    def __init__(
            self, 
            num_queries, 
            num_heads=4, 
            flatten=True, 
            data_format=None, 
            kernel_initializer="orthogonal", 
            kernel_constraint=None, 
            kernel_regularizer=None, 
            name=None, 
            **kwargs
        ):
        super().__init__(
            rank=1, 
            num_queries=num_queries, 
            num_heads=num_heads, 
            flatten=flatten, 
            data_format=data_format, 
            kernel_initializer=kernel_initializer, 
            kernel_constraint=kernel_constraint, 
            kernel_regularizer=kernel_regularizer, 
            name=name, 
            **kwargs
        )


@saving.register_keras_serializable(package="Kerex.layers.pooling", name="AttentionPooling2D")
class AttentionPooling2D(BaseAttentionPooling):
    """
    2D attention based global pooling
    Each query represents a kind of feature or pattern in the input,
    which the input is attended to.

    Parameters
    ----------
    num_queries : int,
        Number of learned queries.
    num_heads : int, optional
        Number of heads in the multihead attention layer.
        Defaults to `4`.
    flatten : bool, optional
        If `True`, the output is flattened over the number of input channels and the number of queries.
        Defaults to `True`.
    kernel_initializer : str | keras.initializer.Initializer, optional
        Initializer for real- and imaginary weights.
        Defaults to `"orthogonal"`.
    kernel_regularizer : str | keras.regularizers.Regularizer, optional
        Kernel regularizer.
        Defaults to `None`.
    kernel_constraint : str | keras.constraints.Constraint, optional
        Kernel constraint.
        Defaults to `None`.
    name : str, optional
        Name of the model.
        If `None`, `name` is automatically inherited from the class name `"BaseSpectralConv"`.
        Defaults to `None`.

    Notes
    -----
    If `flatten=True`, this layer is basically a drop-in replacement for any global pooling layer,
    The output channels depend on the input channels and the number of queries:
    `out_channels = in_channels * num_queries`.

    """
    
    def __init__(
            self, 
            num_queries, 
            num_heads=4, 
            flatten=True, 
            data_format=None, 
            kernel_initializer="orthogonal", 
            kernel_constraint=None, 
            kernel_regularizer=None, 
            name=None, 
            **kwargs
        ):
        super().__init__(
            rank=2, 
            num_queries=num_queries, 
            num_heads=num_heads, 
            flatten=flatten, 
            data_format=data_format, 
            kernel_initializer=kernel_initializer, 
            kernel_constraint=kernel_constraint, 
            kernel_regularizer=kernel_regularizer, 
            name=name, 
            **kwargs
        )


@saving.register_keras_serializable(package="Kerex.layers.pooling", name="AttentionPooling3D")
class AttentionPooling3D(BaseAttentionPooling):
    """
    3D attention based global pooling
    Each query represents a kind of feature or pattern in the input,
    which the input is attended to.

    Parameters
    ----------
    num_queries : int,
        Number of learned queries.
    num_heads : int, optional
        Number of heads in the multihead attention layer.
        Defaults to `4`.
    flatten : bool, optional
        If `True`, the output is flattened over the number of input channels and the number of queries.
        Defaults to `True`.
    kernel_initializer : str | keras.initializer.Initializer, optional
        Initializer for real- and imaginary weights.
        Defaults to `"orthogonal"`.
    kernel_regularizer : str | keras.regularizers.Regularizer, optional
        Kernel regularizer.
        Defaults to `None`.
    kernel_constraint : str | keras.constraints.Constraint, optional
        Kernel constraint.
        Defaults to `None`.
    name : str, optional
        Name of the model.
        If `None`, `name` is automatically inherited from the class name `"BaseSpectralConv"`.
        Defaults to `None`.

    Notes
    -----
    If `flatten=True`, this layer is basically a drop-in replacement for any global pooling layer,
    The output channels depend on the input channels and the number of queries:
    `out_channels = in_channels * num_queries`.

    """
    
    def __init__(
            self, 
            num_queries, 
            num_heads=4, 
            flatten=True, 
            data_format=None, 
            kernel_initializer="orthogonal", 
            kernel_constraint=None, 
            kernel_regularizer=None, 
            name=None, 
            **kwargs
        ):
        super().__init__(
            rank=3, 
            num_queries=num_queries, 
            num_heads=num_heads, 
            flatten=flatten, 
            data_format=data_format, 
            kernel_initializer=kernel_initializer, 
            kernel_constraint=kernel_constraint, 
            kernel_regularizer=kernel_regularizer, 
            name=name, 
            **kwargs
        )



from keras import layers
from keras import Sequential
from keras import ops
from keras.src import activations
from keras.src.backend import standardize_data_format
from ...ops import get_layer


class BaseChannelAttentionModule(layers.Layer):
    """
    Convolutional Channel Attention Module (CAM)

    Parameters
    ----------
    rank : int
        Rank of the convolutions.
    reduction_ratio : int, optional
        Size of the bottleneck of the shared MLP in relation to the input size.
        Defaults to 8.
    activation : str, optional
        Activation function for the shared MLP.
        Defaults to `"sigmoid"`.
    data_format : str, optional {`"channels_first"`, `"channels_last"`}
        Data format of the convolutions.
        Can be either `"channels_first"` or `"channels_last"`.
        Defaults to `"channels_last"`.
    keepdims : bool, optional
        If `True`, the dimensions are preserved (as 1) throughout the pooling process.
        Select `False` to get a global pooling layer that automatically weighs the importance of channels.
        Defaults to `True`.
    name : str, optional
        Name of the layer.
        Defaults to `None`.
    **kwargs : Additional keyword arguments for the `Layer` super-class.

    Notes
    -----
    This is not a wrapper but a CAM layer.
    May be used on it's own, but here it serves as a building block for the 
    [Convolutional Block Attention Module (CBAM)](https://arxiv.org/abs/1807.06521).
    
    More information on the CAM can be found here: https://www.digitalocean.com/community/tutorials/attention-mechanisms-in-computer-vision-cbam#channel-attention-module-cam
    
    """

    def __init__(self, rank, reduction_ratio=8, activation="sigmoid", data_format="channels_last", keepdims=True, name=None, **kwargs):
        super().__init__(name=name, **kwargs)
        self.rank = rank
        self.reduction_ratio = reduction_ratio
        self.activation = activation
        self.data_format = standardize_data_format(data_format)
        self.keepdims = keepdims
        self.channel_axis = -1 if self.data_format == "channels_last" else 1

        self.avg_pooling = get_layer(f"GlobalAveragePooling{self.rank}D", data_format=self.data_format, keepdims=self.keepdims)
        self.max_pooling = get_layer(f"GlobalMaxPooling{self.rank}D", data_format=self.data_format, keepdims=self.keepdims)
        self.merge_layer = layers.Concatenate(axis=self.channel_axis)
        self.activation_fn = activations.get(self.activation)

    def build(self, input_shape):
        """
        Build method of `ChannelAttentionModule`
        
        Builds all sub-modules and sets `self.built=True`.

        Parameters
        ----------
        input_shape : tuple
            Shape of the input.

        """

        if self.built:
            return

        # define mlp
        input_shape = list(input_shape)
        num_ch = input_shape[self.channel_axis]
        self.mlp = Sequential([
            layers.Dense(units=num_ch // self.reduction_ratio, activation="relu"),
            layers.Dense(units=num_ch)
        ])

        # forward pass with updated input shapes
        self.max_pooling.build(input_shape=input_shape)
        self.avg_pooling.build(input_shape=input_shape)

        maxpool_output_shape = self.max_pooling.compute_output_shape(input_shape=input_shape)
        avgpool_output_shape = self.avg_pooling.compute_output_shape(input_shape=input_shape)

        self.merge_layer.build(input_shape=(maxpool_output_shape, avgpool_output_shape))

        merge_output_shape = self.merge_layer.compute_output_shape(input_shape=(maxpool_output_shape, avgpool_output_shape))

        self.mlp.build(input_shape=merge_output_shape)

        self.built = True

    def call(self, inputs):
        """
        Call method of `ChannelAttentionModule`

        Parameters
        ----------
        inputs : KerasTensor
            Input tensor for `ChannelAttentionModule`.

        Returns
        -------
        outputs : KerasTensor
            Refined features.

        """

        max_pool = self.max_pooling(inputs)  # shape (batch, 1, 1, ch)
        avg_pool = self.avg_pooling(inputs)  # shape (batch, 1, 1, ch)

        # concatenate
        x = ops.concatenate([max_pool, avg_pool], axis=self.channel_axis)  # shape = (batch, 1, 1, 2*ch)

        # apply mlp
        x = self.mlp(x)
        x = self.activation_fn(x)

        return x
    
    def compute_output_shape(self, input_shape: tuple) -> tuple:
        """
        Compute output shape

        Parameters
        ----------
        input_shape : tuple
            Shape of the input.

        Returns
        -------
        output_shape : tuple
            Shape of the output of `ChannelAttentionModule`.

        """

        maxpool_output_shape = self.max_pooling.compute_output_shape(input_shape=input_shape)
        avgpool_output_shape = self.avg_pooling.compute_output_shape(input_shape=input_shape)
        merge_output_shape = self.merge_layer.compute_output_shape(input_shape=(maxpool_output_shape, avgpool_output_shape))
        mlp_output_shape = self.mlp.compute_output_shape(input_shape=merge_output_shape)

        return tuple(mlp_output_shape)

    def get_config(self):
        """
        Get config method.
        Required for serialization.

        Returns
        -------
        config : dict
            Dictionary with the configuration of `ChannelAttentionModule`

        """
        
        config: dict = super().get_config()
        config.update({
            "reduction_ratio": self.reduction_ratio,
            "activation": self.activation,
            "data_format": self.data_format,
            "keepdims": self.keepdims
        })
        return config


try:
    import keras
except ImportError as e:
    raise ImportError(
        "The keras package is not installed. It can be installed using 'pip install keras'."
    ) from e

from packaging import version

MIN_KERAS_VERSION = "3.10.0"

if version.parse(keras.__version__) >= version.parse(MIN_KERAS_VERSION):
    from keras.src.layers.convolutional.base_conv import BaseConv
    from keras.src.layers.convolutional.base_conv_transpose import BaseConvTranspose
    from keras.src.layers.convolutional.base_separable_conv import BaseSeparableConv
    from keras.src.layers.convolutional.conv1d import Conv1D
    from keras.src.layers.convolutional.conv1d_transpose import Conv1DTranspose
    from keras.src.layers.convolutional.conv2d import Conv2D
    from keras.src.layers.convolutional.conv2d_transpose import Conv2DTranspose
    from keras.src.layers.convolutional.conv3d import Conv3D
    from keras.src.layers.convolutional.conv3d_transpose import Conv3DTranspose

else:
    """ 
    Keras versions <3.10.0 had faulty serialization code for Convolutional layers,
    which throw an error once the activation function was defined as a layer, e.g.,
    `activation=layers.ReLU(negative_slope=0.1)`, 
    c.f. https://github.com/keras-team/keras/issues/21088.

    The following implementations fix that problem for Keras<3.10.0

    """
    
    from .base_conv import MyBaseConv as BaseConv
    from .base_conv import MyBaseConvTranspose as BaseConvTranspose
    from .base_conv import MyBaseSeparableConv as BaseSeparableConv
    from .conv1d import Conv1D, Conv1DTranspose
    from .conv2d import Conv2D, Conv2DTranspose
    from .conv3d import Conv3D, Conv3DTranspose



from .base_fft_upsampling import BaseFFTUpSampling
from keras import saving


@saving.register_keras_serializable(package="Kerex.Layers.Reshape", name="FFTUpSampling1D")
class FFTUpSampling1D(BaseFFTUpSampling):
    def __init__(
            self, 
            target_size, 
            data_format=None, 
            name=None, 
            **kwargs
        ):
        super().__init__(
            rank=1, 
            target_size=target_size, 
            data_format=data_format, 
            name=name, 
            **kwargs
        )


@saving.register_keras_serializable(package="Kerex.Layers.Reshape", name="FFTUpSampling2D")
class FFTUpSampling2D(BaseFFTUpSampling):
    def __init__(
            self, 
            target_size, 
            data_format=None, 
            name=None, 
            **kwargs
        ):
        super().__init__(
            rank=2, 
            target_size=target_size, 
            data_format=data_format, 
            name=name, 
            **kwargs
        )


@saving.register_keras_serializable(package="Kerex.Layers.Reshape", name="FFTUpSampling3D")
class FFTUpSampling3D(BaseFFTUpSampling):
    def __init__(
            self, 
            target_size, 
            data_format=None, 
            name=None, 
            **kwargs
        ):
        super().__init__(
            rank=3, 
            target_size=target_size, 
            data_format=data_format, 
            name=name, 
            **kwargs
        )


from keras.src.layers.merging.base_merge import Merge
from keras import layers
from keras import KerasTensor
from keras import saving
from ...ops import get_layer


@saving.register_keras_serializable(package="Kerex.Layers.Merge", name="AttentionGate")
class AttentionGate(Merge):
    """
    Attention gate for expressive and expensive merging of two tensors

    Parameters
    ----------
    data_format : str, optional {`"channels_first"`, `"channels_last"`}
        Format of the data. Defaults to `"channels_last"`.
    merge_layer : str | Layer, optional {`"concatenate"`, `"average"`, `"maximum"`, `"minimum"`, `"add"`, `"subtract"`, `"multiply"`, `"dot"`}
        Identifier or layer of a valid Keras merge layer, cf. https://keras.io/api/layers/merging_layers/.
        Defaults to `"concatenate"`.
    **kwargs : Additional keyword arguments for the `Merge` super-class.
    
    Raises
    ------
    ValueError
        If `merge_layer` is not a valid Keras merge layer, cf. https://keras.io/api/layers/merging_layers/

    Notes
    -----
    Original paper: https://arxiv.org/pdf/1804.03999

    `merge="add"` is default additive attention gate mechanism from paper,
    `merge="concatenate"` would allow for different input shapes.

    """

    def __init__(self, data_format="channels_last", merge_layer="add", **kwargs):
        super().__init__(**kwargs)
        self.data_format = data_format
        self.axis = -1 if self.data_format == "channels_last" else 1
        try:
            self.merge_layer = get_layer(merge_layer, axis=self.axis)
        except TypeError:
            # if layer does not support axis keyword try again without
            self.merge_layer = get_layer(merge_layer)

        if not issubclass(type(self.merge_layer), Merge):
            raise ValueError(
                f"Merge-layer {self.merge_layer} supplied to Wrapper isn't "
                "a supported merge-layer."
            )

        self.relu = layers.ReLU()
        self.gate = layers.Multiply()

    def build(self, input_shape):
        """
        Build method of `AttentionGate`

        Parameters
        ----------
        input_shape : tuple
            Shape of the input data.

        """

        # Used purely for shape validation.
        if len(input_shape) < 1 or not isinstance(
            input_shape[0], (tuple, list)
        ):
            raise ValueError(
                "A `AttentionGate` layer should be called on a list of "
                f"at least 1 input. Received: input_shape={input_shape}"
            )
        if all(shape is None for shape in input_shape):
            return
        
        # If the inputs have different ranks, we have to reshape them  # TODO check what this exactly does in Merge
        # to make them broadcastable.
        if None not in input_shape and len(set(map(len, input_shape))) == 1:
            self._reshape_required = False
        else:
            self._reshape_required = True

        # derive dimension for convolution from input shapes
        shape_x, shape_g = input_shape
        
        dim = len(shape_x)
        if dim < 3:
            raise ValueError(
                "Dimension of input data too small,"
                f"expect dim>=3, received dim={dim}."
            )
        
        if dim > 5:
            raise ValueError(
                "Dimension of input data too large,"
                f"expect dim<=5, received dim={dim}."
            )
        
        channels_x = shape_x[self.axis]
        channels_g = shape_g[self.axis]

        # channels = min(channels_x, channels_g)

        conv_kwargs = dict(kernel_size=1, padding="same", data_format=self.data_format)

        # get layers
        self.conv_x = get_layer(f"Conv{dim - 2}D", filters=channels_g, **conv_kwargs)
        self.conv_g = get_layer(f"Conv{dim - 2}D", filters=channels_g, **conv_kwargs)
        self.psi = get_layer(f"Conv{dim - 2}D", filters=channels_x, activation="sigmoid", **conv_kwargs)

        # build layers in forward fashion
        self.conv_x.build(input_shape=shape_x)
        self.conv_g.build(input_shape=shape_g)

        shape_x_forward = self.conv_x.compute_output_shape(input_shape=shape_x)
        shape_g = self.conv_g.compute_output_shape(input_shape=shape_g)

        self.merge_layer.build(input_shape=[shape_x_forward, shape_g])  # NOTE this will throw an exception if shapes don't match, so I don't have to implement it myself :-)
        forward_shape = self.merge_layer.compute_output_shape(input_shape=[shape_x_forward, shape_g])
        
        self.psi.build(input_shape=forward_shape)
        shape_alpha = self.psi.compute_output_shape(input_shape=forward_shape)

        # self.merge_layer.build(input_shape=[shape_x, shape_psi])  # NOTE this will throw an exception if shapes don't match, so I don't have to implement it myself :-)
        assert shape_alpha == shape_x  # has to be given to apply gating mechanism

        # build gate layer (just mutliplication)
        self.gate.build(input_shape=[shape_x, shape_alpha])

        self.built = True

    def _merge_function(self, inputs) -> KerasTensor:
        """ unpack inputs.
        convention for UNet model:
        - inputs[0]: x_forward
        - inputs[1]: skip information

        goal is to modulate the skip information, not the forward information,
        at the end, both information (the modulated skip information and the original forward information) are merged
        """
        x_forward, x_skip = inputs

        # apply 1x1 convolutions to inputs
        x_forward_projection = self.conv_x(x_forward)
        x_skip_projection = self.conv_g(x_skip)

        # add inputs and apply relu activation
        alpha = self.merge_layer((x_forward_projection, x_skip_projection))
        alpha = self.relu(alpha)

        # apply 1x1 convolution with sigmoid to get gating behavior
        alpha = self.psi(alpha)

        # apply gating mechanism to x
        # modulated_skip = x_skip * alpha  # NOTE shapes have to match exactly!
        
        # finally apply the merge operation
        return self.gate([x_forward, alpha])
    
    def get_config(self):
        """
        Get config method.
        Required for serialization.

        Returns
        -------
        config : dict
            Dictionary with the configuration of `AttentionGate`.

        """

        config: dict = super().get_config()
        config.update({
            "data_format": self.data_format,
            "merge_layer": saving.serialize_keras_object(self.merge_layer)
        })
        return config

    @classmethod
    def from_config(cls, config: dict):
        """
        From config method.
        Required for deserialization.

        Parameters
        ----------
        config : dict
            Dictionary with the configuration of `AttentionGate`.

        Returns
        -------
        cls : AttentionGate
            A instance of class `AttentionGate` based on `config`.
            
        """
        
        merge_cfg = config.pop("merge_layer")
        config.update({"merge_layer": saving.deserialize_keras_object(merge_cfg)})
        
        return cls(**config)
    
    def compute_output_shape(self, input_shape):
        """
        Compute output shape

        Parameters
        ----------
        input_shape : tuple
            Shape of the input.

        Returns
        -------
        output_shape : tuple
            Shape of the output of `AttentionGate`.

        """

        x_shape, g_shape = input_shape

        # go through alpha branch 
        alpha_shape = self.merge_layer.compute_output_shape(input_shape=[x_shape, g_shape])
        alpha_shape = self.psi.compute_output_shape(input_shape=alpha_shape)
        
        output_shape = self.gate.compute_output_shape(input_shape=[x_shape, alpha_shape])

        return output_shape
        


import inspect
from keras import ops
from keras import layers
from keras import saving
from keras import KerasTensor
from keras.src import backend
import string
from math import pi
import re


def fftfreq(n, d=1, rad=False):
    """
    Return the Discrete Fourier Transform sample frequencies.

    The returned float array `f` contains the frequency bin centers in cycles
    per unit of the sample spacing (with zero at the start).  For instance, if
    the sample spacing is in seconds, then the frequency unit is cycles/second.

    Given a window length `n` and a sample spacing `d`::

      f = [0, 1, ...,   n/2-1,     -n/2, ..., -1] / (d*n)   if n is even
      f = [0, 1, ..., (n-1)/2, -(n-1)/2, ..., -1] / (d*n)   if n is odd

    Parameters
    ----------
    n : int
        Window length.
    d : scalar, optional
        Sample spacing (inverse of the sampling rate). Defaults to `1`.
    rad : bool, optional
        If this is set, the angular frequency `omega=2*pi*f` is returned.
        Defaults to `False`.

    Returns
    -------
    f : KerasTensor
        Tensor of length `n` containing the sample frequencies.

    Examples
    --------
    >>> from keras import ops
    >>> from ssp.keras.ops import fft, fftfreq
    >>> signal = ops.array([-2, 8, 6, 4, 1, 0, 3, 5], dtype=float)
    >>> fourier = fft(signal)
    >>> n = ops.size(signal)
    >>> timestep = 0.1
    >>> freq = fftfreq(n, d=timestep)
    >>> freq
    array([ 0.  ,  1.25,  2.5 , ..., -3.75, -2.5 , -1.25])

    """

    fs = 1.0 / d
    df = fs / ops.cast(n, float)
    fft_freqs = ops.arange(-ops.cast(n // 2, float) * df, ops.cast(n // 2, float) * df, df)

    if rad:
        fft_freqs *= (2 * pi)

    return ops.roll(fft_freqs, shift=n // 2)


def squeeze_or_expand_to_same_rank(x1, x2, axis=-1, expand_rank_1: bool = True) -> tuple:
    """
    Squeeze/expand along `axis` if ranks differ from expected by exactly 1.

    Parameters
    ----------
    x1 : KerasTensor
        first input tensor
    x2 : KerasTensor
        second input tensor
    axis : int, optional
        axis to squeeze or expand along. Defaults to `-1`.
    expand_rank_1: bool, optional
        Defaults to `True`

    Returns
    -------
    x1, x2 : (KerasTensor, KerasTensor)
        Tuple of `(x1, x2)` with the same shape

    """

    x1_rank = len(x1.shape)
    x2_rank = len(x2.shape)
    if x1_rank == x2_rank:
        return x1, x2
    if x1_rank == x2_rank + 1:
        if x1.shape[axis] == 1:
            if x2_rank == 1 and expand_rank_1:
                x2 = ops.expand_dims(x2, axis=axis)
            else:
                x1 = ops.squeeze(x1, axis=axis)
    if x2_rank == x1_rank + 1:
        if x2.shape[axis] == 1:
            if x1_rank == 1 and expand_rank_1:
                x1 = ops.expand_dims(x1, axis=axis)
            else:
                x2 = ops.squeeze(x2, axis=axis)
    return x1, x2


def large_negative_number(dtype):
    """
    Return a Large negative number based on dtype.

    Parameters
    ----------
    dtype : str
        dtype of large negative number to return
    
    Returns
    -------
    c : float
        Large negative number with dtype `dtype` (-1e9 for `dtype="float32"`, -3e4 for `dtype=float16"`).
    
    """
    
    if backend.standardize_dtype(dtype) == "float16":
        return -3e4
    return -1e9


def index_to_einsum_variable(i):
    """Coverts an index to a einsum variable name.

    We simply map indices to lowercase characters, e.g. 0 -> 'a', 1 -> 'b'.

    """

    return string.ascii_lowercase[i]


def unwrap(phase, axis=-1, period=2*pi):
    """
    Unwrap by taking the complement of large deltas with respect to the period.
    Inspired by https://numpy.org/doc/stable/reference/generated/numpy.unwrap.html

    Parameters
    ----------
    phase : KerasTensor
        Input array.
    axis : int, optional
        Axis along which unwrap will operater.
        Defaults to the last axis.
    period : float, optional
        Size of the range over which the input wraps. By default, it is
        `2*pi`.

    Returns
    -------
    out : KerasTensor
        Output array.

    Examples
    --------
    >>> from keras import ops
    >>> from math import pi
    >>> phase = ops.linspace(0, pi, 5) + ops.array([0.0, 0.0, 0.0, pi, pi])
    >>> ops.convert_to_numpy(phase)
    array([0.       , 0.7853982, 1.5707964, 5.4977875, 6.2831855], dtype=float32)
    >>> ops.convert_to_numpy(unwrap(phase))
    array([ 0.      , 0.7853982, 1.5707964, -0.785398, 0.       ], dtype=float32)

    """

    nd = ops.ndim(phase)

    # get phase difference and correction
    phase_diff = ops.diff(phase, axis=axis)
    jumps = ops.cast(phase_diff < -period / 2, dtype="int8") - ops.cast(phase_diff > period / 2, dtype="int8")
    correction = ops.cumsum(ops.cast(jumps, dtype="float32") * period, axis=axis)

    # pad to original size
    pad_width = [(0, 0)] * nd
    pad_width[axis] = (1, 0)

    correction = ops.pad(correction, pad_width=tuple(pad_width))
    return phase + correction


def capitalize_first_char(s):
    """
    Capitalize first character of string and leave the rest.
    

    Parameters
    ----------
    s : str
        A string to process.

    Returns
    -------
    s : str
        A modified string where the initial character is capitalized and the rest remains original.

    Notes
    -----
    Beneficial to import layers, e.g., "Conv2D".capitalize() results in "Conv2d", which is no valid layer!
    Source: https://stackoverflow.com/questions/12410242/python-capitalize-first-letter-only
    
    """

    return re.sub('([a-zA-Z])', lambda x: x.groups()[0].upper(), s, 1)


@saving.register_keras_serializable(package="Kerex.Helper", name="get_layer")
def get_layer(identifier, module="keras.layers", registered_name=None, **layer_kwargs):
    """
    Get a layer from an identifier

    Parameters
    ----------
    identifier : dict | str | Layer
        The identifier for a layer to return.
    module : str, optional
        The module to load the layer from.
        When loading official Keras layer, this should be `"keras.layers"`.
        When using a custom package, this should be the path to this package, e.g., `"keras_addon.layers"`.
        Defauts to `"keras.layers"`.
    registered_name : str, optional
        Only necessary if loading inofficial keras layers. Defaults to `None`.

    Returns
    -------
    layer : Layer
        An instance of a keras.Layer class from the identifier.

    """

    if identifier is None:
        return None
    if isinstance(identifier, dict):
        obj = layers.deserialize(identifier)
    elif isinstance(identifier, str):
        config = {
            "module": module,
            "class_name": str(capitalize_first_char(identifier)),  # layer names are all capital!
            "config": {
                "name": None,
                "trainable": True,
                "dtype": {
                    "module": "keras",
                    "class_name": "DTypePolicy",
                    "config": {"name": "float32"},
                    "registered_name": None
                },
                **layer_kwargs
            },
            "registered_name": registered_name
        }
        obj = layers.deserialize(config)
    else:
        obj = identifier

    if callable(obj):
        if inspect.isclass(obj):
            obj = obj()
        return obj
    else:
        raise ValueError(
            f"Could not interpret layer identifier: {identifier}"
        )


class _IterableVars:
    """
    This class just provides basic functionality for convolutional blocks

    **Use for inheritance only!**

    This class provides a function to set the class variables as iterables based on the the first keyword argument.

    """

    def set_vars(self, **kwargs):
        """
        Wrap all `**kwargs` to list and set as class attribute

        The first keyword argument has a special role here, as it determines the depth of the network.
        All other parameters (if given as a list or tuple initially) have to comply with the length of this keyword argument.

        Parameters
        ----------
        **kwargs : Keyword arguments.

        Raises
        ------
        ValueError
            If an argument is a list/tuple and contains other dtypes than `int`, `float`, `str`, `bool` or `tuple`.
        ValueError
            If an argument is a list/tuple and contains tuples that do not match `self.rank`.
            This error is only raised when a model has a class attribute `self.rank`.
        ValueError
            If an argument does not match the length of the reference key.
        
        """

        if not kwargs:
            return

        # get the reference key, which has to be the first keyword
        ref_key = list(kwargs.keys())[0]  # e.g., `filters`

        # now iterate over kwargs
        for k, arg in kwargs.items():
            if isinstance(arg, (list, tuple)):
                if len(arg) == 1:
                    # unpack list/tuple of length 1
                    arg, = arg

            if isinstance(arg, (int, float, str, bool, tuple)) or (arg is None):
                if hasattr(self, ref_key):
                    # wrap singular argument (that is not `filters` in list)
                    arg = [arg] * len(getattr(self, ref_key))
                else:
                    # wrap `ref_key` argument in list if it is not a tuple (iterable) already
                    if not isinstance(arg, tuple):
                        arg = [arg]
            
            # wrap arguments that come as nested list (happens after deserialization!) to a list of tuples
            arg = [tuple(f) if isinstance(f, list) else f for f in arg]

            if not all(isinstance(f, (int, float, str, bool, tuple)) or (f is None) for f in arg):
                raise ValueError(f"Received bad `{k}` argument ({arg}). Expected all entries of `{k}` to be either `int`, `str`, `bool`, or `tuple`.")
            
            if hasattr(self, ref_key):
                if hasattr(self, "rank"):
                    if not all(len(f) == self.rank for f in arg if isinstance(f, tuple)):
                        raise ValueError(f"Rank of provided `{ref_key}` does not match rank of the block or model, expected `{k}` of rank {self.rank}.")
            
            try:
                if not len(arg) == len(getattr(self, ref_key)):
                    raise ValueError(f"Too many arguments for `{k}`, expected {len(getattr(self, ref_key))}, received {len(arg)}.")
            except AttributeError:
                # `ref_key` was not set yet, no comparison possible
                pass
            
            # set `v` as class attribute `k`
            setattr(self, k, arg)



from .base_encoder import BaseEncoder
from keras import saving


@saving.register_keras_serializable(package="Kerex.Blocks.Autoencoder", name="Encoder1D")
class Encoder1D(BaseEncoder):
    """
    1-D convolutional encoder block

    Parameters
    ----------
    filters : int | list | tuple
        Number of filters for the convolutional forward sub-model.
    kernel_size : int | list | tuple, optional
        Kernel size for the convolutional forward sub-model.
        Defaults to 5.
    strides : int | list | tuple, optional
        Strides for the convolutional forward sub-model.
        Defaults to 1.
    padding : str, optional {`"same"`, `"causal"`}
        Padding that is applied to maintain deterministic data shapes.
        Defaults to `"same"`.
    data_format : str, optional {`"channels_first"`, `"channels_last"`}
        Data format for the convolution operations.
        Defaults to `None`.
    dilation_rate : int | list | tuple, optional
        Dilation rate for the convolutional forward sub-model.
        Defaults to 1.
    groups : int | list | tuple, optional
        Number of groups rate for the convolutional forward sub-model.
        Defaults to 1.
    downsampling_filters : int, optional
        Number of filters for the downsampling operation.
        If `None`, this parameter is set to the last entry of `filters`.
        Defaults to `None`.
    downsampling_groups : int, optional
        Number of groups for the downsampling operation.
        If `None`, this parameter is set to the last entry of `filters`.
        Defaults to `None`.
    activation : str | keras.activations.Activation | keras.layers.Layer, optional
        Activation for the convolutional forward sub-model.
        Can be either a `str`, a `keras.activations.Activation`, or a `keras.layers.Layer`.
        Defaults to `"relu"`.
    use_skip_connection : bool, optional
        Whether to use skip connections.
        Defaults to `False`.
    use_bias : bool, optional
        Whether to use bias.
        Defaults to `True`.
    kernel_initializer : str | keras.initializers.Initializer, optional
        Kernel initializer.
        Defaults to `"he_normal"`.
    bias_initializer : str | keras.initializers.Initializer, optional
        Bias initializer.
        Defaults to `"zeros"`.
    kernel_regularizer : str | keras.regularizers.Regularizer, optional
        Kernel regularizer.
        Defaults to `None`.
    bias_regularizer : str | keras.regularizers.Regularizer, optional
        Bias regularizer.
        Defaults to `None`.
    kernel_constraint : str | keras.constraints.Constraint, optional
        Kernel constraint.
        Defaults to `None`.
    bias_constraint : str | keras.constraints.Constraint, optional
        Bias constraint.
        Defaults to `None`.
    name : str, optional
        Name of the layer.
        If `None`, `name` is automatically inherited from the class name `"Encoder1D"`.
        Defaults to `None`.
    **kwargs : Additional keyword arguments for the `keras.layers.Layer` super-class.

    """

    def __init__(
            self, 
            filters, 
            kernel_size=5, 
            strides=1, 
            padding="same",
            data_format=None, 
            dilation_rate=1, 
            groups=1, 
            use_skip_connection = False, 
            activation="relu", 
            use_bias=True, 
            kernel_initializer="he_normal", 
            bias_initializer="zeros", 
            kernel_regularizer=None, 
            bias_regularizer=None, 
            kernel_constraint=None, 
            bias_constraint=None, 
            name=None, 
            **kwargs
        ):
        assert padding in ["same", "causal"], f"Allowed padding types for `Encoder1D` are `'same'` and `'causal'`, received {padding}."
        super().__init__(
            rank=1, 
            filters=filters, 
            kernel_size=kernel_size, 
            strides=strides, 
            padding=padding,
            data_format=data_format, 
            dilation_rate=dilation_rate, 
            groups=groups, 
            use_skip_connection=use_skip_connection, 
            activation=activation, 
            use_bias=use_bias, 
            kernel_initializer=kernel_initializer, 
            bias_initializer=bias_initializer, 
            kernel_regularizer=kernel_regularizer, 
            bias_regularizer=bias_regularizer, 
            kernel_constraint=kernel_constraint, 
            bias_constraint=bias_constraint, 
            name=name, 
            **kwargs
        )
    

@saving.register_keras_serializable(package="Kerex.Blocks.Autoencoder", name="Encoder2D")
class Encoder2D(BaseEncoder):
    """
    2-D convolutional encoder block

    Parameters
    ----------
    filters : int | list | tuple
        Number of filters for the convolutional forward sub-model.
    kernel_size : int | list | tuple, optional
        Kernel size for the convolutional forward sub-model.
        Defaults to 5.
    strides : int | list | tuple, optional
        Strides for the convolutional forward sub-model.
        Defaults to 1.
    padding : str, optional {`"same"`, `"causal"`}
        Padding that is applied to maintain deterministic data shapes.
        Defaults to `"same"`.
    data_format : str, optional {`"channels_first"`, `"channels_last"`}
        Data format for the convolution operations.
        Defaults to `None`.
    dilation_rate : int | list | tuple, optional
        Dilation rate for the convolutional forward sub-model.
        Defaults to 1.
    groups : int | list | tuple, optional
        Number of groups rate for the convolutional forward sub-model.
        Defaults to 1.
    downsampling_filters : int, optional
        Number of filters for the downsampling operation.
        If `None`, this parameter is set to the last entry of `filters`.
        Defaults to `None`.
    downsampling_groups : int, optional
        Number of groups for the downsampling operation.
        If `None`, this parameter is set to the last entry of `filters`.
        Defaults to `None`.
    activation : str | keras.activations.Activation | keras.layers.Layer, optional
        Activation for the convolutional forward sub-model.
        Can be either a `str`, a `keras.activations.Activation`, or a `keras.layers.Layer`.
        Defaults to `"relu"`.
    use_skip_connection : bool, optional
        Whether to use skip connections.
        Defaults to `False`.
    use_bias : bool, optional
        Whether to use bias.
        Defaults to `True`.
    kernel_initializer : str | keras.initializers.Initializer, optional
        Kernel initializer.
        Defaults to `"he_normal"`.
    bias_initializer : str | keras.initializers.Initializer, optional
        Bias initializer.
        Defaults to `"zeros"`.
    kernel_regularizer : str | keras.regularizers.Regularizer, optional
        Kernel regularizer.
        Defaults to `None`.
    bias_regularizer : str | keras.regularizers.Regularizer, optional
        Bias regularizer.
        Defaults to `None`.
    kernel_constraint : str | keras.constraints.Constraint, optional
        Kernel constraint.
        Defaults to `None`.
    bias_constraint : str | keras.constraints.Constraint, optional
        Bias constraint.
        Defaults to `None`.
    name : str, optional
        Name of the layer.
        If `None`, `name` is automatically inherited from the class name `"Encoder2D"`.
        Defaults to `None`.
    **kwargs : Additional keyword arguments for the `keras.layers.Layer` super-class.

    """

    def __init__(
            self, 
            filters, 
            kernel_size=5, 
            strides=1, 
            padding="same", 
            data_format=None, 
            dilation_rate=1, 
            groups=1, 
            use_skip_connection = False, 
            activation="relu", 
            use_bias=True, 
            kernel_initializer="he_normal", 
            bias_initializer="zeros", 
            kernel_regularizer=None, 
            bias_regularizer=None, 
            kernel_constraint=None, 
            bias_constraint=None, 
            name=None, 
            **kwargs
        ):
        super().__init__(
            rank=2, 
            filters=filters, 
            kernel_size=kernel_size, 
            strides=strides, 
            padding=padding, 
            data_format=data_format, 
            dilation_rate=dilation_rate, 
            groups=groups, 
            use_skip_connection=use_skip_connection, 
            activation=activation, 
            use_bias=use_bias, 
            kernel_initializer=kernel_initializer, 
            bias_initializer=bias_initializer, 
            kernel_regularizer=kernel_regularizer, 
            bias_regularizer=bias_regularizer, 
            kernel_constraint=kernel_constraint, 
            bias_constraint=bias_constraint, 
            name=name, 
            **kwargs
        )


@saving.register_keras_serializable(package="Kerex.Blocks.Autoencoder", name="Encoder3D")
class Encoder3D(BaseEncoder):
    """
    3-D convolutional encoder block

    Parameters
    ----------
    filters : int | list | tuple
        Number of filters for the convolutional forward sub-model.
    kernel_size : int | list | tuple, optional
        Kernel size for the convolutional forward sub-model.
        Defaults to 5.
    strides : int | list | tuple, optional
        Strides for the convolutional forward sub-model.
        Defaults to 1.
    padding : str, optional {`"same"`, `"causal"`}
        Padding that is applied to maintain deterministic data shapes.
        Defaults to `"same"`.
    data_format : str, optional {`"channels_first"`, `"channels_last"`}
        Data format for the convolution operations.
        Defaults to `None`.
    dilation_rate : int | list | tuple, optional
        Dilation rate for the convolutional forward sub-model.
        Defaults to 1.
    groups : int | list | tuple, optional
        Number of groups rate for the convolutional forward sub-model.
        Defaults to 1.
    downsampling_filters : int, optional
        Number of filters for the downsampling operation.
        If `None`, this parameter is set to the last entry of `filters`.
        Defaults to `None`.
    downsampling_groups : int, optional
        Number of groups for the downsampling operation.
        If `None`, this parameter is set to the last entry of `filters`.
        Defaults to `None`.
    activation : str | keras.activations.Activation | keras.layers.Layer, optional
        Activation for the convolutional forward sub-model.
        Can be either a `str`, a `keras.activations.Activation`, or a `keras.layers.Layer`.
        Defaults to `"relu"`.
    use_skip_connection : bool, optional
        Whether to use skip connections.
        Defaults to `False`.
    use_bias : bool, optional
        Whether to use bias.
        Defaults to `True`.
    kernel_initializer : str | keras.initializers.Initializer, optional
        Kernel initializer.
        Defaults to `"he_normal"`.
    bias_initializer : str | keras.initializers.Initializer, optional
        Bias initializer.
        Defaults to `"zeros"`.
    kernel_regularizer : str | keras.regularizers.Regularizer, optional
        Kernel regularizer.
        Defaults to `None`.
    bias_regularizer : str | keras.regularizers.Regularizer, optional
        Bias regularizer.
        Defaults to `None`.
    kernel_constraint : str | keras.constraints.Constraint, optional
        Kernel constraint.
        Defaults to `None`.
    bias_constraint : str | keras.constraints.Constraint, optional
        Bias constraint.
        Defaults to `None`.
    name : str, optional
        Name of the layer.
        If `None`, `name` is automatically inherited from the class name `"Encoder3D"`.
        Defaults to `None`.
    **kwargs : Additional keyword arguments for the `keras.layers.Layer` super-class.

    """

    def __init__(
            self, 
            filters, 
            kernel_size=5, 
            strides=1, 
            padding="same", 
            data_format=None, 
            dilation_rate=1, 
            groups=1, 
            use_skip_connection = False, 
            activation="relu", 
            use_bias=True, 
            kernel_initializer="he_normal", 
            bias_initializer="zeros", 
            kernel_regularizer=None, 
            bias_regularizer=None, 
            kernel_constraint=None, 
            bias_constraint=None, 
            name=None, 
            **kwargs
        ):
        super().__init__(
            rank=3, 
            filters=filters, 
            kernel_size=kernel_size, 
            strides=strides, 
            padding=padding, 
            data_format=data_format, 
            dilation_rate=dilation_rate, 
            groups=groups, 
            use_skip_connection=use_skip_connection, 
            activation=activation, 
            use_bias=use_bias, 
            kernel_initializer=kernel_initializer, 
            bias_initializer=bias_initializer, 
            kernel_regularizer=kernel_regularizer, 
            bias_regularizer=bias_regularizer, 
            kernel_constraint=kernel_constraint, 
            bias_constraint=bias_constraint, 
            name=name, 
            **kwargs
        )



from keras import layers
from keras import ops
from keras import saving
from keras.src.models import Sequential
from keras.src.models import Functional
from keras.src.layers.convolutional.base_conv import BaseConv
from keras.src.layers.convolutional.base_separable_conv import BaseSeparableConv
from keras.src.layers.convolutional.base_conv_transpose import BaseConvTranspose
from keras.src.backend import standardize_data_format


@saving.register_keras_serializable(package="Kerex.Layers.Wrapper", name="TemporalSlice")
class TemporalSlice(layers.Wrapper):
    """
    Temporal slice wrapper

    Can be used to wrap any Keras `layer` or `Model`.
    This wrapper executes the wrapped Keras `layer` or `Model` multiple times and stacks the output.

    For example, given an input `(1, 32, 10)` of shape `(batch_size, x, t)`,
    this wrapper can be used to wrap a pre-trained model which processes inputs of shape `(1, 32, 8)`.
    With `strides=1`, the model has to be called three times in order to process the input of shape `(1, 32, 10)`.
    
    Parameters
    ----------
    layer : Layer
        Keras `layer` or `Model` to be wrapped.
    window_size : int
        Length (time steps) of the input to a single wrapped `layer` or `Model`.
    strides : int, optional
        Strides between temporal windows.
        Defaults to 1.
    axis : int, optional
        Axis along which the sequence of outputs is stacked along.
        If not set, the axis is derived from the `data_format`.
        Defaults to `None`.
    name : str, optional
        Name of the wrapper, that is prepended to the name of the wrapped layer.
        Defaults to `"temporal_slice"`.
    **kwargs : Additional keyword arguments for the `Wrapper` super-class.

    Raises
    ------
    RuntimeError
        If there is a reshape layer in the `Model` that is not the last layer.

    Notes
    -----
    The layer is fully compatible with Keras automatic differentiation,
    and thus can be used on trainable layers and Models.

    If the wrapped layer is a Sequential model which contains a (non-trainable) reshape layer,
    the layer is removed, since it complicates things a lot!

    """

    def __init__(self, layer, window_size, strides=1, axis=None, name="temporal_slice", **kwargs):
        name = "_".join([name, layer.name])
        super().__init__(layer, **kwargs)

        if isinstance(self.layer, Sequential):
            is_reshape_layer = [issubclass(type(layer), layers.Reshape) for layer in self.layer.layers]
            if any(is_reshape_layer):
                if (sum(is_reshape_layer) == 1) and (is_reshape_layer.index(True) + 1 == len(is_reshape_layer)):
                    # there is only a single reshape layer and it is the last layer in the Sequential model. we can simply remove it
                    model_layers: list = self.layer.layers
                    model_layers.pop(-1)
                    self.layer = Sequential(model_layers)
                else:
                    raise RuntimeError(f"Wrapped model has reshape layers. Please remove those before wrapping model.")

        # inherit data format from wrapped layer
        try:
            self.data_format = self.layer.data_format
            self.is_convolutional = True
        except AttributeError:
            if isinstance(self.layer, Sequential):
                seq_layers = self.layer.layers
            elif not isinstance(self.layer, Functional):
                # We treat subclassed models as a simple sequence of layers, for logging
                # purposes.
                seq_layers = self.layer.layers

            conv_layers = [isinstance(type(l), (BaseConv, BaseConvTranspose, BaseSeparableConv)) for l in seq_layers]
            if any(conv_layers):
                self.is_convolutional = True
                self.data_format = seq_layers[conv_layers.index(True)].data_format
            else:
                self.is_convolutional = False
                self.data_format = "channels_last"  # make an educated guess regarding the data format

        # set axis for temporal slice. defaults to the channel axis / -1
        self.axis = axis or -1 if self.data_format == "channels_last" else 1

        # sliding window parameters
        self.window_size = window_size
        self.strides = strides
        self.num_windows = None  # is set in build method, since it depends on the input shape!

        self.transpose_axes = None  # is set in build method, too

    def build(self, input_shape):
        """
        Build method of `TemporalSlice`.

        Builds all sub-modules and sets `self.built=True`.

        Parameters
        ----------
        input_shape : tuple
            Shape of the input data.

        """

        # get feature axes, i.e., axes that remain untouched for input shape for a single wrapped instance
        feature_axes = list(range(len(input_shape)))
        feature_axes.pop(0)  # remove batch
        feature_axes.pop(self.axis)  # remove axis we operate on

        # update input shape
        input_slice = [None, *[input_shape[item] for item in feature_axes]]

        # if self.is_convolutional:
        if self.axis == -1:
            input_slice.append(self.window_size)
        else:
            input_slice.insert(self.axis, self.window_size)

        input_slice = tuple(input_slice)

        # build super()
        super().build(input_shape=input_slice)

        # now, set sliding window params
        sequence_length = input_shape[self.axis]
        self.num_windows = max(0, (sequence_length - self.window_size) // self.strides + 1)
        if self.num_windows < 2:
            raise ValueError(f"Invalid number of windows {self.num_windows}. Increase window size ({self.window_size}) or increase strides ({self.strides}).")

        """ define transpose axis
        The output of the vectorized map will be (self.num_windows, *output_shape)
        We first have to move the first dimension to the last axis,
        to then stack the last two axes with a reshape layer
        """
        # define transpose axes
        transpose_axes = list(range(1, len(input_shape) + 1 if self.is_convolutional else len(input_shape)))

        # first, move first axis (0) after the original channel axis
        if self.axis == -1:
            transpose_axes.append(0)
        else:
            transpose_axes.insert(self.axis + 1 if self.is_convolutional else self.axis, 0)

        self.transpose_axes = transpose_axes

        """ define reshape op
        we have to merge the channel axes
        """
        # if self.is_convolutional:
        output_shape = self.layer.compute_output_shape(input_shape=input_slice)

        # remove batch dimension from output shape
        output_shape = list(output_shape)
        output_shape.pop(0)

        target_shape = [1] * len(output_shape)
        target_shape[self.axis] = self.num_windows

        self.target_shape = tuple([int(ts * os) for ts, os in zip(output_shape, target_shape)])

        self.reshape = layers.Reshape(target_shape=self.target_shape)

        self.built = True

    def call(self, inputs):
        """
        Call method of `TemporalSlice`

        The input is first processed using a sliding window along `self.axis` of the data.
        Then, the wrapped `layer` or `Model` is called on each slice.
        The outputs are stacked along `self.axis` of the data.

        Parameters
        ----------
        inputs : KerasTensor
            Input tensor for the wrapped Keras `layer` or `Model`.

        Returns
        -------
        outputs : KerasTensor
            Output tensor of the wrapped Keras `layer` or `Model`.

        """

        slices = ops.array([ops.take(inputs, indices=ops.arange(i, i + self.window_size, self.strides), axis=self.axis) for i in range(self.num_windows)])        
        x = ops.vectorized_map(self.layer, slices)

        # combine first and last dimension somehow. we need to stack them ideally
        x = ops.transpose(x, axes=self.transpose_axes)
        if self.is_convolutional:
            x = self.reshape(x)

        return x
    
    def get_config(self):
        """
        Get config method.
        Required for serialization.

        Returns
        -------
        config : dict
            Dictionary with the configuration of `Residual`.

        """

        config: dict = super().get_config()
        config.update({
            "window_size": self.window_size,
            "strides": self.strides,
            "axis": self.axis
        })
        return config
    
    @classmethod
    def from_config(cls, config):
        """
        From config method.
        Required for deserialization.

        Parameters
        ----------
        config : dict
            Dictionary with the configuration of `TemporalSlice`.

        Returns
        -------
        cls : TemporalSlice
            A instance of class `TemporalSlice` based on `config`.

        """

        layer_config = config.pop('layer')
        layer = saving.deserialize_keras_object(layer_config)

        return cls(layer=layer, **config)

    def compute_output_shape(self, input_shape):
        """
        Compute output shape

        Parameters
        ----------
        input_shape : tuple
            Shape of the input.

        Returns
        -------
        output_shape : tuple
            Shape of the output of `TemporalSlice`.

        """

        input_shape = list(input_shape)
        b = input_shape.pop(0)
        output_shape = (b, *self.target_shape)
        return output_shape


class TemporalSliceV2(layers.Wrapper):
    """
    call model on windows along an axis of the data

    e.g. 
    input_shape = (1, 16, 32, 32, 3)
    model_input_shape = (1, 4, 32, 32, 3)
    model_output_shape = (1, 1, 32, 32, 3)
    --> call model 13 times along first axis of input and stacks the output along the same axis
    --> TemporalSliceV2(model): (1, 16, 32, 32, 3) -> (1, 13, 32, 32, 3)

    It is assumed, that the wrapper layer or model is convolutional

    """

    def __init__(self, layer, window_size, strides=1, axis=None, data_format=None, name="temporal_slice", **kwargs):
        name = f"{name}({layer.name})"
        super().__init__(layer=layer, name=name, **kwargs)

        # remove reshape layer from any `Sequential` model
        if isinstance(self.layer, Sequential):
            is_reshape_layer = [issubclass(type(layer), layers.Reshape) for layer in self.layer.layers]
            if any(is_reshape_layer):
                if (sum(is_reshape_layer) == 1) and (is_reshape_layer.index(True) + 1 == len(is_reshape_layer)):
                    # there is only a single reshape layer and it is the last layer in the Sequential model. we can simply remove it
                    model_layers: list = self.layer.layers
                    model_layers.pop(-1)
                    self.layer = Sequential(model_layers)
                else:
                    raise RuntimeError(f"Wrapped model has reshape layers. Please remove those before wrapping model.")

        self.data_format = standardize_data_format(data_format)  # defaults to `"channels_last"`

        # set axis for temporal slice. defaults to the channel axis / -1
        self.axis = axis or -1 if self.data_format == "channels_last" else 1

        # sliding window parameters
        self.window_size = window_size
        self.strides = strides
        self.num_windows = None  # is set in build method, since it depends on the input shape!

        self.transpose_axes = None  # is set in build method, too

    def build(self, input_shape):
        if self.built:
            return
        
        # get feature axes, i.e., axes that remain untouched for input shape for a single wrapped instance
        feature_axes = list(range(len(input_shape)))
        feature_axes.pop(0)  # remove batch
        feature_axes.pop(self.axis)  # remove axis we operate on

        # update input shape
        input_slice = [None, *[input_shape[item] for item in feature_axes]]

        if self.axis == -1:
            input_slice.append(self.window_size)
        else:
            input_slice.insert(self.axis, self.window_size)

        input_slice = tuple(input_slice)

        # build super()
        super().build(input_shape=input_slice)

        # define input_spec
        channel_axis = input_shape[self.axis]
        input_time_steps = input_shape[self.axis]

        self.input_spec = layers.InputSpec(ndim=len(input_slice) - 2, axes={channel_axis: input_time_steps})


from. autoencoder import FCN1D, FCN2D, FCN3D, SmoothFCN1D, SmoothFCN2D, SmoothFCN3D, FourierFCN1D, FourierFCN2D, FourierFCN3D
from .autoencoder import Unet1D, Unet2D, Unet3D, SmoothUnet1D, SmoothUnet2D, SmoothUnet3D, FourierUnet1D, FourierUnet2D, FourierUnet3D

from .neural_operator import NeuralOperator1D, NeuralOperator2D, NeuralOperator3D

from .tsmixer import TSMixer


"""
Since the deserialization is broken in the keras BaseConv when using a layer (layers.ReLU()) as activation function,
I had to implement my own base conv and inherit ConvXD from it.

A simple fix in the get_config() and from_config() methods allows to use both sting identifiers ("relu") and layers (layers.ReLU()) as activation function
"""
from keras.src.layers.convolutional.base_conv import BaseConv
from keras.src.layers.convolutional.base_conv_transpose import BaseConvTranspose
from keras.src.layers.convolutional.base_separable_conv import BaseSeparableConv
from keras import saving


class MyBaseConv(BaseConv):
    def get_config(self):
        config: dict = super().get_config()
        config.update({"activation": saving.serialize_keras_object(self.activation)})

        return config
    
    @classmethod
    def from_config(cls, config: dict):
        activation_cfg = config.pop("activation")
        config.update({"activation": saving.deserialize_keras_object(activation_cfg)})

        return cls(**config)
    

class MyBaseConvTranspose(BaseConvTranspose):
    def get_config(self):
        config: dict = super().get_config()
        config.update({"activation": saving.serialize_keras_object(self.activation)})

        return config
    
    @classmethod
    def from_config(cls, config: dict):
        activation_cfg = config.pop("activation")
        config.update({"activation": saving.deserialize_keras_object(activation_cfg)})

        return cls(**config)
    

class MyBaseSeparableConv(BaseSeparableConv):
    def get_config(self):
        config: dict = super().get_config()
        config.update({"activation": saving.serialize_keras_object(self.activation)})

        return config
    
    @classmethod
    def from_config(cls, config: dict):
        activation_cfg = config.pop("activation")
        config.update({"activation": saving.deserialize_keras_object(activation_cfg)})

        return cls(**config)
    


from keras import layers, Layer
from keras import InputSpec
from keras.src.backend import standardize_data_format
from keras import initializers
from keras import regularizers
from keras import constraints
from keras import ops


class BaseAttentionPooling(Layer):
    """
    Base layer for attention based global pooling
    Each query represents a kind of feature or pattern in the input,
    which the input is attended to.

    Parameters
    ----------
    rank : int {1, 2, 3}
        Rank of the subclassed layer
    num_queries : int,
        Number of learned queries.
    num_heads : int, optional
        Number of heads in the multihead attention layer.
        Defaults to `4`.
    flatten : bool, optional
        If `True`, the output is flattened over the number of input channels and the number of queries.
        Defaults to `True`.
    kernel_initializer : str | keras.initializer.Initializer, optional
        Initializer for real- and imaginary weights.
        Defaults to `"orthogonal"`.
    kernel_regularizer : str | keras.regularizers.Regularizer, optional
        Kernel regularizer.
        Defaults to `None`.
    kernel_constraint : str | keras.constraints.Constraint, optional
        Kernel constraint.
        Defaults to `None`.
    name : str, optional
        Name of the model.
        If `None`, `name` is automatically inherited from the class name `"BaseSpectralConv"`.
        Defaults to `None`.

    Notes
    -----
    If `flatten=True`, this layer is basically a drop-in replacement for any global pooling layer,
    The output channels depend on the input channels and the number of queries:
    `out_channels = in_channels * num_queries`.

    """
    
    def __init__(
            self, 
            rank,
            num_queries, 
            num_heads=4,
            flatten=True,
            data_format=None,
            kernel_initializer="orthogonal",
            kernel_constraint=None,
            kernel_regularizer=None,
            name=None, 
            **kwargs
        ):
        super().__init__(
            name=name,
            **kwargs
        )
        self.rank = rank
        self.num_queries = num_queries
        self.num_heads = num_heads
        
        self.flatten = flatten

        self.data_format = standardize_data_format(data_format)

        self.kernel_initializer = initializers.get(kernel_initializer)
        self.kernel_constraint = constraints.get(kernel_constraint)
        self.kernel_regularizer = regularizers.get(kernel_regularizer)

        if self.flatten:
            self.flattening_layer = layers.Flatten(data_format=self.data_format)

    def build(self, input_shape):
        if self.built:
            return

        if self.data_format == "channels_last":
            channel_axis = -1
            d_emb = input_shape[-1]
            target_shape = [-1, d_emb]

        else:
            channel_axis = 1
            d_emb = input_shape[1]
            target_shape = [d_emb, -1]

        # reshape layer, flattens input if necessary
        self.reshape = layers.Reshape(target_shape=target_shape)
        self.reshape.build(input_shape)

        # define input spec
        self.input_spec = InputSpec(
            min_ndim=self.rank + 2, axes={channel_axis: d_emb}
        )
        
        # add learnable query vector
        self.query = self.add_weight(
            name="learnable_queries",
            shape=(1, self.num_queries, d_emb),
            initializer=self.kernel_initializer,
            regularizer=self.kernel_regularizer,
            constraint=self.kernel_constraint,
            trainable=True,
            dtype=self.dtype
        )

        # now define and build MHA layer
        self.mha = layers.MultiHeadAttention(
            num_heads=self.num_heads,
            key_dim=d_emb
        )
        mha_input_shape = self.reshape.compute_output_shape(input_shape)
        self.mha.build(query_shape=(None, self.num_queries, d_emb), value_shape=mha_input_shape)

        self.built = True

    def call(self, inputs):
        flat_inputs = self.reshape(inputs)

        # tile queries across batch
        b = ops.shape(inputs)[0]
        q = ops.tile(self.query, [b, 1, 1])

        pooled = self.mha(query=q, value=flat_inputs)

        if self.flatten:
            pooled = self.flattening_layer(pooled)

        return pooled
    
    def compute_output_shape(self, input_shape):
        b = input_shape[0]
        d_emb = input_shape[-1 if self.data_format == "channels_last" else 1]

        flat_input_shape = self.reshape.compute_output_shape(input_shape)

        pooled_shape = self.mha.compute_output_shape(
            query_shape=tuple([b, self.num_queries, d_emb]),
            value_shape=flat_input_shape
        )

        if self.flatten:
            pooled_shape = self.flattening_layer.compute_output_shape(pooled_shape)

        return pooled_shape
    
    def get_config(self):
        config = super().get_config()
        config.update({
            "num_queries": self.num_queries,
            "num_heads": self.num_heads,
            "flatten": self.flatten,
            "data_format": self.data_format,
            "kernel_initializer": initializers.serialize(self.kernel_initializer),
            "kernel_constraint": constraints.serialize(self.kernel_constraint),
            "kernel_regularizer": regularizers.serialize(self.kernel_regularizer)
        })

        return config
    
    @classmethod
    def from_config(cls, config):
        kernel_initializer_cfg = config.pop("kernel_initializer")
        kernel_constraint_cfg = config.pop("kernel_constraint")
        kernel_regularizer_cfg = config.pop("kernel_regularizer")

        config.update({
            "kernel_initializer": initializers.deserialize(kernel_initializer_cfg),
            "kernel_constraint": constraints.deserialize(kernel_constraint_cfg),
            "kernel_regularizer": regularizers.deserialize(kernel_regularizer_cfg)
        })

        return cls(**config)
    


from keras import layers
from keras import saving
from keras import ops
from functools import partial
from typing import Tuple
from ...blocks.attention import GlobalSelfAttention


@saving.register_keras_serializable(package="Kerex.Layers.Wrapper", name="SpatialAttention")
class SpatialAttention(layers.Wrapper):
    """
    Applies spatial multi-head attention on the wrapped layer.
    
    Parameters
    ----------
    layer : Layer
        Keras 'layer' to be wrapped
    num_heads : int, optional
        Number of attention heads. Defaults to 4.
    key_dim : int, optional
        Size of key dimension of multi-head attention. Defaults to 64.
    value_dim : int, optional
        Size of value dimension of multi-head attention. Defaults to 64.
    dropout : float, optional
        Dropout propability. Defaults to 0.
    name : str
        Name of the wrapper.
        Is prepended to the name of the wrapped layer.
        Defaults to `"spatial_attention"`.
    **kwargs : Additional keyword arguments for the `Wrapper` super-class.

    Raises
    ------
    ValueError
        If wrapped `layer` is `LSTM` with argument `return_sequences=False`.

    Notes
    -----
    This layer does not add too many weights to the model,
    but is very expensive to run.

    The implementation follows https://arxiv.org/abs/2307.09072

    """

    def __init__(self, layer, num_heads=4, key_dim=64, value_dim=64, dropout=0, name="spatial_attention", **kwargs):
        if issubclass(type(layer), layers.LSTM):
            if not layer.return_sequences:
                raise ValueError(f"LSTM layer has to be initialized with 'return_sequences=True' in order to work with this wrapper!")
        
        name = "_".join([layer.name, name])
        super().__init__(layer=layer, name=name, **kwargs)
        self.num_heads = num_heads
        self.key_dim = key_dim
        self.value_dim = value_dim
        self.dropout = dropout
        try:
            self.data_format = self.layer.data_format
        except AttributeError:
            self.data_format = "channels_last"  # this is valid for dense layer for example! TODO: check if it really is ;-D

        self.channel_axis = -1 if self.data_format == "channels_last" else 1

        self.self_attention = GlobalSelfAttention(
            num_heads=num_heads,
            key_dim=key_dim,
            value_dim=value_dim,
            dropout=dropout
        )

    def get_transpose(self, feature_axes: list, input_shape: tuple) -> Tuple[list]:
        # we have to transpose to 'channels_last' for attention! Target transpose axes = [0, *feature_axes, 1]
        # N: sequence length from features
        # d_model: number of channels
        if self.data_format == "channels_first":
            # construct transpose axes from feature axes. For attention, we need data format [b, N, d_model] where N is the sequence length
            transpose_axes = [0, *feature_axes, 1]  # for example [0, 2, 1]
            inverse_transpose_axes = [0, len(input_shape) - 1, *[fa - 1 for fa in feature_axes]]  # ...has to be [0, 2, 1]
        else:
            transpose_axes = list(range(len(input_shape)))
            inverse_transpose_axes = transpose_axes.copy()

        return transpose_axes, inverse_transpose_axes
    
    def get_reshape(self, feature_axes: list, output_shape: tuple) -> Tuple[int]:
        # now we have to reshape in case len(feature_axis) < 1
        batch_size = output_shape[0]
        feature_shape = [output_shape[a] for a in feature_axes]

        sequence_length = int(ops.prod(feature_shape))
        num_channel = output_shape[self.channel_axis]

        newshape = (batch_size, sequence_length, num_channel)
        inverse_newshape = (batch_size, *feature_shape, num_channel)
        
        return newshape, inverse_newshape

    def build(self, input_shape: tuple):
        """
        Build method of `SpatialAttention`.

        Builds all sub-modules and sets `self.built=True`.

        Parameters
        ----------
        input_shape : tuple
            Shape of the input data.

        """

        super().build(input_shape=input_shape)
        output_shape = self.layer.compute_output_shape(input_shape)
        
        # extract feature axes from input shape
        feature_axes = list(range(len(input_shape)))
        feature_axes.pop(self.channel_axis)  # remove channel dimension, e.g. [1] when channels last or self.channel_axis = -1
        feature_axes.pop(0)  # remove batch dimension  # e.g. [1, 2]

        # get transpose operation
        transpose_axes, inverse_transpose_axes = self.get_transpose(feature_axes=feature_axes, input_shape=input_shape)

        self.transpose = partial(ops.transpose, axes=transpose_axes)
        self.inverse_transpose = partial(ops.transpose, axes=inverse_transpose_axes)

        # get reshape operation
        newshape, inverse_newshape = self.get_reshape(feature_axes=feature_axes, output_shape=output_shape)

        # both newshape and inverse_newshape have (None) as first dimension. For the transpose operation, we change that to (-1)
        self.reshape = partial(ops.reshape, newshape=(-1, *newshape[1:]))
        self.inverse_reshape = partial(ops.reshape, newshape=(-1, *inverse_newshape[1:]))

        # finally, build attention layer
        self.self_attention.build(query_shape=newshape, value_shape=newshape)

    def call(self, inputs):
        """
        Call method of `SpatialAttention`

        Parameters
        ----------
        inputs : KerasTensor
            Input tensor for the wrapped Keras `layer` or `Model`.

        Returns
        -------
        outputs : KerasTensor
            Output tensor of the wrapped Keras `layer` or `Model`.

        """

        x = self.layer(inputs)

        # apply attention to input
        attention_inputs = self.transpose(x)
        attention_inputs = self.reshape(x)
        attention_outputs = self.self_attention(attention_inputs)

        # inverse transform
        attention_outputs = self.inverse_reshape(attention_outputs)
        attention_outputs = self.inverse_transpose(attention_outputs)
        return x + attention_outputs

    def get_config(self):
        """
        Get config method.
        Required for serialization.

        Returns
        -------
        config : dict
            Dictionary with the configuration of `SpatialAttention`.

        """
        
        config: dict = super().get_config()
        config.update({
            "num_heads": self.num_heads,
            "key_dim": self.key_dim,
            "value_dim": self.value_dim,
            "dropout": self.dropout
        })
        return config

    def compute_output_shape(self, input_shape):
        """
        Compute output shape

        Parameters
        ----------
        input_shape : tuple
            Shape of the input.

        Returns
        -------
        output_shape : tuple
            Shape of the output of `SpatialAttention`.

        """

        return self.layer.compute_output_shape(input_shape)
    

@saving.register_keras_serializable(package="Kerex.Layers.Wrapper", name="ChannelAttention")
class ChannelAttention(SpatialAttention):
    """
    Applies channel-wise multi-head attention on the wrapped layer.
    
    Parameters
    ----------
    layer : Layer
        Layer to be wrapped
    num_heads : int, optional
        Number of attention heads. Defaults to 4.
    key_dim : int, optional
        Size of key dimension of multi-head attention. Defaults to 64.
    value_dim : int, optional
        Size of value dimension of multi-head attention. Defaults to 64.
    dropout : float, optional
        Dropout propability. Defaults to 0.
    name : str
        Name of the wrapper.
        Is prepended to the name of the wrapped layer.
        Defaults to `"channel_attention"`.
    **kwargs : Additional keyword arguments for the `Wrapper` super-class.

    Raises
    ------
    ValueError
        If wrapped `layer` is `LSTM` with argument `return_sequences=False`.

    Notes
    -----
    This layer adds a lot of weights to the model,
    but is still more lightweight at runtime than the `SpatialAttention` wrapper.

    The implementation follows https://arxiv.org/pdf/2112.13047

    """

    def __init__(self, layer, num_heads=4, key_dim=64, value_dim=64, dropout=0, name="channel_attention", **kwargs):
        super().__init__(
            layer=layer,
            num_heads=num_heads,
            key_dim=key_dim,
            value_dim=value_dim,
            dropout=dropout,
            name=name,
            **kwargs
        )

    def get_transpose(self, feature_axes: list, input_shape: Tuple) -> Tuple[list]:
        # we have to transpose to 'channels_first' for attention! Target transpose axes = [0, 1, *feature_axes]
        if self.data_format == "channels_last":
            # construct transpose axes from feature axes. For attention, we need data format [b, N, d_model] where N is the sequence length
            # N: number of channels
            # d_model: sequence length from features
            transpose_axes = [0, len(input_shape) - 1, *feature_axes]
            inverse_transpose_axes = [0, *[fa + 1 for fa in feature_axes], 1]
        else:
            transpose_axes = list(range(len(input_shape)))
            inverse_transpose_axes = transpose_axes.copy()

        return transpose_axes, inverse_transpose_axes
    
    def get_reshape(self, feature_axes: list, output_shape: tuple) -> Tuple[int]:
        # now we have to reshape in case len(feature_axis) < 1
        batch_size = output_shape[0]
        feature_shape = [output_shape[a] for a in feature_axes]
        num_channels = int(ops.prod(feature_shape))
        sequence_length = output_shape[self.channel_axis]

        newshape = (batch_size, sequence_length, num_channels)
        newshape_inverse = (batch_size, sequence_length, *feature_shape)

        return newshape, newshape_inverse



from keras import layers
from keras import Sequential
from keras import initializers
from keras import saving
from keras import ops


@saving.register_keras_serializable(package="Kerex.Layers.Wrapper", name="FiLM")
class FiLM(layers.Wrapper):
    """
    Feature-wise Linear Modulation Wrapper

    This class provides a wrapper that allows to condition any Keras layer to an additional input.

    Parameters
    ----------
    layer : Layer
        Keras 'layer' or `Model` to be wrapped.
    num_hidden : int, optional
        Size of latent space of FiLM sub-network.
        Defaults to 16.
    name : str, optional
        Name of the wrapper, that is prepended to the name of the wrapped layer.
        Defaults to `"film_conditioned"`.
    **kwargs : Additional keyword arguments for the `Wrapper` super-class.

    Notes
    -----
    The FiLM approach generates two additional biases, `gamma` and `beta`.
    The activation of the layer `layer(x)` is refined by `y = gamma * layer(x) + beta`.
    The FiLM sub-networks are initiallized such that `gamma~1` and `beta~0`.
    
    The layer refines the features following the [FiLM](https://arxiv.org/abs/1709.07871) approach.

    """

    def __init__(self, layer, num_hidden=16, name="film_conditioned", **kwargs):
        name = "_".join([name, layer.name])
        super().__init__(layer=layer, name=name, **kwargs)
        self.num_hidden = num_hidden

        # inherit information from wrapped layer
        try:
            self.data_format = self.layer.data_format
        except AttributeError:
            self.data_format = "channels_last"  # this is valid for dense layer for example! TODO: check if it really is ;-D

        self.beta = None
        self.gamma = None

    def build(self, input_shape, conditioning_input_shape: tuple):
        """
        Build method of `FiLM`
        
        Builds all sub-modules and sets `self.built=True`.

        Parameters
        ----------
        input_shape : tuple
            Shape of the input.
        conditioning_input_shape : tuple
            Shape of the input used to conidtion the wrapped layer.

        """

        self.layer.build(input_shape)
        output_shape = self.layer.compute_output_shape(input_shape)

        # the output shape of FiLM is a tensor with shape (num_ch,)
        # it's basically and additional bias term!
        num_ch = output_shape[-1 if self.data_format == "channels_last" else 1]

        self.beta = Sequential([
            layers.Dense(
                units=self.num_hidden, 
                activation="relu"
            ),
            layers.Dense(
                units=num_ch, 
                activation="linear", 
                kernel_initializer=initializers.RandomNormal(mean=0.0, stddev=0.01),
                bias_initializer=initializers.Zeros()
            )
        ])
        self.gamma = Sequential([
            layers.Dense(
                units=self.num_hidden, 
                activation="relu"
            ),
            layers.Dense(
                units=num_ch, 
                activation="sigmoid", 
                kernel_initializer=initializers.RandomNormal(mean=0.0, stddev=0.01),
                bias_initializer=initializers.Zeros()
            ),
            layers.Lambda(lambda x: 0.75 + x * 0.5)  # rescale activation. With this, gamma is always in [0.75, 1.25]. Optionally, try x + 0.5 for symmetric activation between [0.5, 1.5]
        ])

        # now build the two sub-networks. For this, we need their input shape
        self.beta.build(conditioning_input_shape)
        self.gamma.build(conditioning_input_shape)

        # declare einsum operation for multiplication
        self.einsum_ops = "bc...,bc->bc..." if self.data_format == "channels_first" else "b...c,bc->b...c"

    def call(self, inputs, conditioning_inputs):
        """
        Call method of `FiLM`.

        Parameters
        ----------
        inputs : KerasTensor
            Input tensor for `FiLM`.
        conditioning_inputs : KerasTensor
            Conditioning input for `FiLM`

        Returns
        -------
        outputs : KerasTensor
            Refined features.

        """

        x = self.layer(inputs)
        beta = self.beta(conditioning_inputs)
        gamma = self.gamma(conditioning_inputs)

        # multiply using einsum
        y1 = ops.einsum(self.einsum_ops, x, gamma)

        # add bias; first, broadcast to same shape as x with einsum!
        bias = ops.einsum(self.einsum_ops, ops.ones_like(x), beta)
        
        return y1 + bias
    
    def get_config(self):
        """
        Get config method.
        Required for serialization.

        Returns
        -------
        config : dict
            Dictionary with the configuration of `SpatialAttentionModule`

        """

        config: dict = super().get_config()
        config.update({"num_hidden": self.num_hidden})
        return config
    
    def compute_output_shape(self, input_shape: tuple, conditioning_input_shape: tuple) -> tuple:
        """
        Compute output shape

        Parameters
        ----------
        input_shape : tuple
            Shape of the input.
        conditioning_input_shape : tuple
            Shape of the conditioning input.

        Returns
        -------
        output_shape : tuple
            Shape of the output of `CBAM`.

        Notes
        -----
        The `conditioning_input` has no impact on the shape.
        In general, `FiLM` does not modify the output shape at all.
        Hence, this layer simply calls the `compute_output_shape` method of the wrapped layer.

        """

        return self.layer.compute_output_shape(input_shape)
    


from keras import layers
from importlib import import_module


class BaseSmoothUpSampling(layers.Layer):
    def __init__(
        self, 
        rank,
        filters, 
        kernel_size, 
        strides=1,
        padding="valid",
        data_format=None,
        dilation_rate=1,
        groups=1,
        activation=None,
        use_bias=True,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        name=None, 
        **kwargs
    ):
        super().__init__(name=name, **kwargs)
        self.rank = rank
        
        # get layers
        self.upsampling = None  # is set in subclassed layers since the args are vastly different across 1-D, 2-D, and 3-D
        self.conv = getattr(import_module(name="...layers.conv", package=__package__), f"Conv{self.rank}D")(
            filters=filters,
            kernel_size=kernel_size,
            strides=strides,
            padding=padding,
            data_format=data_format,
            dilation_rate=dilation_rate,
            groups=groups,
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            **kwargs
        )

    def build(self, input_shape):
        if self.built:
            return
        
        # define input spec
        channel_axis = len(input_shape) - 1 if self.conv.data_format == "channels_last" else 1
        input_channel = input_shape[channel_axis]
        self.input_spec = layers.InputSpec(ndim=self.rank + 2, axes={channel_axis: input_channel})

        # build upsampling layer
        self.upsampling.build(input_shape)
        input_shape = self.upsampling.compute_output_shape(input_shape=input_shape)

        # build conv layer
        self.conv.build(input_shape=input_shape)

        self.built = True

    def call(self, inputs):
        x = self.upsampling(inputs)
        return self.conv(x)
    
    def compute_output_shape(self, input_shape):
        output_shape = self.upsampling.compute_output_shape(input_shape=input_shape)
        output_shape = self.conv.compute_output_shape(input_shape=output_shape)

        return output_shape
    
    def get_config(self):
        config = super().get_config()

        config.update({**self.upsampling.get_config()})
        config.update({**self.conv.get_config()})

        return config



from keras import layers
from keras import Sequential
from keras import initializers
from keras import regularizers
from keras import constraints
from keras import saving
from ...ops.helper import _IterableVars
from importlib import import_module
from keras.src.backend import standardize_data_format


class BaseEncoder(layers.Layer, _IterableVars):
    """
    Base class of convolutional encoder block

    Use to subclass 1-D, 2-D, and 3-D Encoder

    Parameters
    ----------
    rank : int {1, 2, 3}
        Rank of `BaseEncoder`. Must be within {1, 2, 3}.
    filters : int | list | tuple
        Number of filters for the convolutional forward sub-model.
    kernel_size : int | list | tuple, optional
        Kernel size for the convolutional forward sub-model.
        Defaults to 5.
    strides : int | list | tuple, optional
        Strides for the convolutional forward sub-model.
        Defaults to 1.
    padding : str, optional
        Padding that is applied to maintain deterministic data shapes.
        Defaults to `"same"`.
    data_format : str, optional {`"channels_first"`, `"channels_last"`}
        Data format for the convolution operations.
        Defaults to `"channels_last"`.
    dilation_rate : int | list | tuple, optional
        Dilation rate for the convolutional forward sub-model.
        Defaults to 1.
    groups : int | list | tuple, optional
        Number of groups rate for the convolutional forward sub-model.
        Defaults to 1.
    downsampling_filters : int, optional
        Number of filters for the downsampling operation.
        If `None`, this parameter is set to the last entry of `filters`.
        Defaults to `None`.
    downsampling_groups : int, optional
        Number of groups for the downsampling operation.
        If `None`, this parameter is set to the last entry of `filters`.
        Defaults to `None`.
    activation : str | keras.activations.Activation | keras.layers.Layer, optional
        Activation for the convolutional forward sub-model.
        Can be either a `str`, a `keras.activations.Activation`, or a `keras.layers.Layer`.
        Defaults to `"relu"`.
    use_skip_connection : bool, optional
        Whether to use skip connections.
        Defaults to `False`.
    use_bias : bool, optional
        Whether to use bias.
        Defaults to `True`.
    kernel_initializer : str | keras.initializers.Initializer, optional
        Kernel initializer.
        Defaults to `"he_normal"`.
    bias_initializer : str | keras.initializers.Initializer, optional
        Bias initializer.
        Defaults to `"zeros"`.
    kernel_regularizer : str | keras.regularizers.Regularizer, optional
        Kernel regularizer.
        Defaults to `None`.
    bias_regularizer : str | keras.regularizers.Regularizer, optional
        Bias regularizer.
        Defaults to `None`.
    kernel_constraint : str | keras.constraints.Constraint, optional
        Kernel constraint.
        Defaults to `None`.
    bias_constraint : str | keras.constraints.Constraint, optional
        Bias constraint.
        Defaults to `None`.
    name : str, optional
        Name of the layer.
        If `None`, `name` is automatically inherited from the class name `"BaseEncoder"`.
        Defaults to `None`.
    **kwargs : Additional keyword arguments for the `keras.layers.Layer` super-class.

    Notes
    -----
    Downsampling is realized using a strided convolution as proposed by [Springenberger et al.](https://arxiv.org/abs/1412.6806)

    """

    def __init__(
            self,
            rank,
            filters,
            kernel_size=5,
            strides=1,
            padding="same",
            data_format="channels_last",
            dilation_rate=1,
            groups=1,
            use_skip_connection=False,
            downsampling_filters=None,
            downsampling_groups=None,
            activation="relu",
            use_bias=True,
            kernel_initializer="he_normal",
            bias_initializer="zeros",
            kernel_regularizer=None,
            bias_regularizer=None,
            kernel_constraint=None,
            bias_constraint=None,
            name=None, 
            **kwargs):
        super().__init__(name=name, **kwargs)

        # set general class variables
        self.rank = rank
        self.padding = padding
        self.data_format = standardize_data_format(data_format)
        self.activation = activation
        self.use_skip_connection = use_skip_connection
        self.use_bias = use_bias
        self.kernel_initializer = kernel_initializer
        self.bias_initializer = bias_initializer
        self.kernel_regularizer = kernel_regularizer
        self.bias_regularizer = bias_regularizer
        self.kernel_constraint = kernel_constraint
        self.bias_constraint = bias_constraint

        # set class variables for the Sequential model
        self.set_vars(filters=filters, kernel_size=kernel_size, strides=strides, dilation_rate=dilation_rate, groups=groups)

        # set class variables that are relevant for downsampling operation
        self.downsampling_filters = downsampling_filters or self.filters[-1]
        self.downsampling_groups = downsampling_groups or self.groups[-1]

        # define layers
        self.forward_conv = Sequential([
            getattr(import_module(name="...layers.conv", package=__package__), f"Conv{self.rank}D")(
                filters=f,
                kernel_size=k,
                strides=s,
                padding=self.padding,
                data_format=self.data_format,
                dilation_rate=d,
                groups=g,
                activation=self.activation,
                use_bias=self.use_bias,
                kernel_initializer=self.kernel_initializer,
                bias_initializer=self.bias_initializer,
                kernel_regularizer=self.kernel_regularizer,
                bias_regularizer=self.bias_regularizer,
                activity_regularizer=self.activity_regularizer,
                kernel_constraint=self.kernel_constraint,
                bias_constraint=self.bias_constraint
            ) for f, k, s, d, g in zip(self.filters, self.kernel_size, self.strides, self.dilation_rate, self.groups)
        ])

        self.downsampling = getattr(import_module(name="...layers.conv", package=__package__), f"Conv{self.rank}D")(
            filters=self.downsampling_filters,
            kernel_size=2,
            strides=2,
            padding="same",
            data_format=self.data_format,
            dilation_rate=1,
            groups=self.downsampling_groups,
            activation=None,
            use_bias=self.use_bias,
            kernel_initializer=self.kernel_initializer,
            bias_initializer=self.bias_initializer,
            kernel_regularizer=self.kernel_regularizer,
            bias_regularizer=self.bias_regularizer,
            activity_regularizer=self.activity_regularizer,
            kernel_constraint=self.kernel_constraint,
            bias_constraint=self.bias_constraint
        )

    def call(self, inputs):
        """
        Call method of `BaseEncoder`

        The `inputs` are first processed by `self.forward_conv`,
        and then bisected along the feature axes by `self.downsampling`.
        If `self.use_skip_connection=True`, the layer additionally returns the output of `self.forward_conv`
        alongside the output of `self.downsampling`.
        
        Parameters
        ----------
        inputs : KerasTensor
            Input tensor.

        Returns
        -------
        outputs | (outputs_forward, outputs_skip) : KerasTensor | (KerasTensor, KerasTensor)
            Output of `self.downsampling` if `self.use_skip_connection=False`.
            If `self.use_skip_connection=True`, the layer additionally returns the intermediate output before the downsampling operation.

        """

        x_skip = self.forward_conv(inputs)
        x_forward = self.downsampling(x_skip)

        if self.use_skip_connection:
            return x_forward, x_skip
        
        return x_forward
    
    def build(self, input_shape):
        """
        Build method of `BaseEncoder`

        Builds all sub-modules and sets `self.built=True`.
        
        Parameters
        ----------
        input_shape : tuple
            Input shape

        """

        if self.built:
            return
        
        super().build(input_shape)

        # build forward layer and update input_shape
        self.forward_conv.build(input_shape=input_shape)
        input_shape = self.forward_conv.compute_output_shape(input_shape=input_shape)

        # build downsampling layer
        self.downsampling.build(input_shape=input_shape)

        # update built state
        self.built = True
    
    def compute_output_shape(self, input_shape):
        """
        Compute output shape of `BaseEncoder`

        Parameters
        ----------
        input_shape : tuple
            Input shape.

        Returns
        -------
        output_shape : tuple
            Output shape.

        """

        output_shape_skip = self.forward_conv.compute_output_shape(input_shape=input_shape)
        output_shape_forward = self.downsampling.compute_output_shape(input_shape=output_shape_skip)

        if self.use_skip_connection:
            return (output_shape_forward, output_shape_skip)
        
        return output_shape_forward

    def get_config(self):
        """
        Necessary for Keras serialization

        Returns
        -------
        config : dict
            Dictionary with the layer configuration.
            
        """

        config: dict = super().get_config()
        config.update({
            "filters": self.filters,
            "kernel_size": self.kernel_size,
            "strides": self.strides,
            "padding": self.padding,
            "data_format": self.data_format,
            "dilation_rate": self.dilation_rate,
            "groups": self.groups,
            "use_skip_connection": self.use_skip_connection,
            "downsampling_filters": self.downsampling_filters,
            "downsampling_groups": self.downsampling_groups,
            "activation": saving.serialize_keras_object(self.activation),
            "use_bias": self.use_bias,
            "kernel_initializer": initializers.serialize(self.kernel_initializer),
            "bias_initializer": initializers.serialize(self.bias_initializer),
            "kernel_regularizer": regularizers.serialize(self.kernel_regularizer),
            "bias_regularizer": regularizers.serialize(self.bias_regularizer),
            "kernel_constraint": constraints.serialize(self.kernel_constraint),
            "bias_constraint": constraints.serialize(self.bias_constraint)
        })
        
        return config
    
    @classmethod
    def from_config(cls, config):
        """
        Necessary for Keras deserialization

        Parameters
        ----------
        cls : BaseEncoder
            The `BaseEncoder` class.
        config : dict
            Dictionary with the layer configuration.

        Returns
        -------
        cls : BaseEncoder
            Instance of `BaseEncoder` from `config`.
            
        """

        # get configs of keras objects
        activation_cfg = config.pop("activation")
        kernel_initializer_cfg = config.pop("kernel_initializer")
        bias_initializer_cfg = config.pop("bias_initializer")
        kernel_regularizer_cfg = config.pop("kernel_regularizer")
        bias_regularizer_cfg = config.pop("bias_regularizer")
        kernel_constraint_cfg = config.pop("kernel_constraint")
        bias_constraint_cfg = config.pop("bias_constraint")

        config.update({
            "activation": saving.deserialize_keras_object(activation_cfg),
            "kernel_initializer": initializers.deserialize(kernel_initializer_cfg),
            "bias_initializer": initializers.deserialize(bias_initializer_cfg),
            "kernel_regularizer": regularizers.deserialize(kernel_regularizer_cfg),
            "bias_regularizer": regularizers.deserialize(bias_regularizer_cfg),
            "kernel_constraint": constraints.deserialize(kernel_constraint_cfg),
            "bias_constraint": constraints.deserialize(bias_constraint_cfg)
        })

        return cls(**config)
    


from .encoder import Encoder1D, Encoder2D, Encoder3D
from .decoder import Decoder1D, Decoder2D, Decoder3D, SmoothDecoder1D, SmoothDecoder2D, SmoothDecoder3D, FourierDecoder1D, FourierDecoder2D, FourierDecoder3D



from keras import layers
from keras import activations
from keras import initializers
from keras import regularizers
from keras import constraints
from keras import saving
from ...layers.wrapper import Residual
from importlib import import_module


@saving.register_keras_serializable(package="Kerex.layers.FNO", name="FNOInitializer")
class FNOInitializer(initializers.VarianceScaling):
    def __init__(self, seed=None):
        super().__init__(
            scale=0.5,
            mode="fan_avg",
            distribution="truncated_normal",
            seed=seed
        )

    def get_config(self):
        return {
            "seed": saving.serialize_keras_object(self._init_seed)
        }


class BaseFNO(layers.Layer):
    """
    Base FNO layer, cf. https://arxiv.org/abs/2010.08895

    This layer has two paths:

    The first path applies a spectral convolution to the input, which
    (1) transforms the layer into Fourier space via discrete Fourier transform for real-valued data (rfft),
    (2) truncates the Fourier modes to the `modes` lowest modes,
    (3) applies the weights (separated into real- and imaginary weights),
    (4) pads the truncated signal in Fourier space back to its initial shape,
    (5) applies inverse discrete Fourier transform for real-valued data (irfft), and
    (6) applies the bias in physical space (if `use_bias=True`)

    The second path applies a 1x1 bypass convolution

    Both path are merged with `merge_layer` and passed through an activation function.    

    Parameters
    ----------
    rank : int {1, 2, 3}
        Rank of the subclassed FNO layer.
    filters : int
        Number of filters.
    modes : int
        Number of modes after truncation in Fourier space.
    activation : str | keras.activations.Activation | keras.layers.Layer, optional
        Activation function.
        Defaults to `"gelu"`.
    merge_layer : str | keras.layers.Layer, optional {`"concatenate"`, `"average"`, `"maximum"`, `"minimum"`, `"add"`, `"subtract"`, `"multiply"`, `"dot"`}
        Merge operation in FNO layers to combine the result from the spectral convolution with the result from the bypass convolution.
        Defaults to `"add"`.
    data_format : str, optional {`"channels_first"`, `"channels_last"`}
        Format of the input data.
        Defaults to `None`.
    use_bias : bool, optional
        If `True`, a bias term is added to the physical space after the spectral convolution.
        Defaults to `True`.
    kernel_initializer : str | keras.initializer.Initializer | tuple, optional
        Initializer for real- and imaginary weights.
        By default, the real weights are initialized using `"glorot_normal"`,
        and the imaginary weights are initialized using `"random"` with a low standard deviation,
        which is effectively white Gaussian noise.
        Defaults to `("glorot_normal", initializers.RandomNormal(stddev=1e-3))`.
    bias_initializer : str | keras.initializer.Initializer, optional
        Initializer for the bias.
        Defaults to `"zeros"`.
    kernel_regularizer : str | keras.regularizers.Regularizer, optional
        Kernel regularizer.
        Defaults to `None`.
    bias_regularizer : str | keras.regularizers.Regularizer, optional
        Bias regularizer.
        Defaults to `None`.
    kernel_constraint : str | keras.constraints.Constraint, optional
        Kernel constraint.
        Defaults to `None`.
    bias_constraint : str | keras.constraints.Constraint, optional
        Bias constraint.
        Defaults to `None`.
    name : str, optional
        Name of the model.
        If `None`, `name` is automatically inherited from the class name `"SpectralConv1D"`.
        Defaults to `None`.

    Notes
    -----
    For implementataion simplicity, the Fourier operations are always performed in `"channels_first"` data format.
    The layer therefore applies a transpose operation if `data_format="channels_last"`.
    
    """

    def __init__(
        self,
        rank,
        filters,
        modes,
        activation="gelu",
        merge_layer="add",
        data_format=None,
        use_bias=True,
        kernel_initializer=FNOInitializer(),
        bias_initializer="zeros",
        kernel_constraint=None,
        bias_constraint=None,
        kernel_regularizer=None,
        bias_regularizer=None,
        name=None,
        **kwargs
    ):
        super().__init__(name=name, **kwargs)
        self.rank = rank
        self.filters = filters
        self.modes = modes
        self.merge_layer = merge_layer
        self.data_format = data_format
        self.use_bias = use_bias

        if isinstance(kernel_initializer, (list, tuple)):
            self.real_kernel_initializer, self.imag_kernel_initializer = kernel_initializer
        else:
            self.real_kernel_initializer = self.imag_kernel_initializer = kernel_initializer

        self.bias_initializer = bias_initializer
        self.kernel_constraint = kernel_constraint
        self.bias_constraint = bias_constraint
        self.kernel_regularizer = kernel_regularizer
        self.bias_regularizer = bias_regularizer

        self.forward = Residual(
            layer=getattr(import_module(name=".spectral_conv", package=__package__), f"SpectralConv{self.rank}D")(
                filters=self.filters,
                modes=self.modes,
                data_format=self.data_format,
                use_bias=self.use_bias,
                kernel_initializer=(
                    self.real_kernel_initializer,
                    self.imag_kernel_initializer
                ),
                bias_initializer=self.bias_initializer,
                kernel_constraint=self.kernel_constraint,
                bias_constraint=self.bias_constraint,
                kernel_regularizer=self.kernel_regularizer,
                bias_regularizer=self.bias_regularizer,
                name="spectral_conv",
                **kwargs
            ),
            merge_layer=self.merge_layer,
            residual_layer=getattr(import_module(name="...layers.conv", package=__package__), f"Conv{self.rank}D")(
                filters=self.filters,
                kernel_size=1,
                data_format=self.data_format,
                use_bias=self.use_bias,
                kernel_initializer=self.real_kernel_initializer,
                bias_initializer=self.bias_initializer,
                kernel_constraint=self.kernel_constraint,
                bias_constraint=self.bias_constraint,
                kernel_regularizer=self.kernel_regularizer,
                bias_regularizer=self.bias_regularizer,
                name="bypass_conv"
            )
        )
        
        self.activation = activations.get(activation)

    def build(self, input_shape):
        if self.built:
            return
        
        self.forward.build(input_shape=input_shape)

        # inherit input spec from `spectral_conv` layer
        self.input_spec = self.forward.layer.input_spec
        
        self.built = True

    def call(self, inputs):
        x = self.forward(inputs)
        x = self.activation(x)

        return x
    
    def compute_output_shape(self, input_shape):
        """
        Compute output shape of `BaseNeuralOperator`

        Parameters
        ----------
        input_shape : tuple
            Input shape.

        Returns
        -------
        output_shape : tuple
            Output shape.

        """
        
        return self.forward.compute_output_shape(input_shape=input_shape)        
    
    def get_config(self):
        config = super().get_config()
        config.update({
            "filters": self.filters,
            "modes": self.modes,
            "activation": saving.serialize_keras_object(self.activation),
            "merge_layer": saving.serialize_keras_object(self.merge_layer),
            "data_format": self.data_format,
            "use_bias": self.use_bias,
            "real_kernel_initializer": initializers.serialize(self.real_kernel_initializer),
            "imag_kernel_initializer": initializers.serialize(self.imag_kernel_initializer),
            "bias_initializer": initializers.serialize(self.bias_initializer),
            "kernel_constraint": constraints.serialize(self.kernel_constraint),
            "bias_constraint": constraints.serialize(self.bias_constraint),
            "kernel_regularizer": regularizers.serialize(self.kernel_regularizer),
            "bias_regularizer": regularizers.serialize(self.bias_regularizer)
        })

        return config
    
    @classmethod
    def from_config(cls, config):
        activation_cfg = config.pop("activation")
        merge_layer_cfg = config.pop("merge_layer")
        real_kernel_initializer_cfg = config.pop("real_kernel_initializer")
        imag_kernel_initializer_cfg = config.pop("imag_kernel_initializer")
        bias_initializer_cfg = config.pop("bias_initializer")
        kernel_constraint_cfg = config.pop("kernel_constraint")
        bias_constraint_cfg = config.pop("bias_constraint")
        kernel_regularizer_cfg = config.pop("kernel_regularizer")
        bias_regularizer_cfg = config.pop("bias_regularizer")

        config.update({
            "activation": saving.deserialize_keras_object(activation_cfg),
            "merge_layer": saving.deserialize_keras_object(merge_layer_cfg),
            "kernel_initializer": (
                initializers.deserialize(real_kernel_initializer_cfg), 
                initializers.deserialize(imag_kernel_initializer_cfg)
            ),
            "bias_initializer": initializers.deserialize(bias_initializer_cfg),
            "kernel_constraint": constraints.deserialize(kernel_constraint_cfg),
            "bias_constraint": constraints.deserialize(bias_constraint_cfg),
            "kernel_regularizer": regularizers.deserialize(kernel_regularizer_cfg),
            "bias_regularizer": regularizers.deserialize(bias_regularizer_cfg)
        })

        return cls(**config)
    


from keras import saving
from keras import initializers
from .base_neural_operator import BaseNeuralOperator, FNOInitializer


@saving.register_keras_serializable(package="Kerex.Models.NeuralOperator", name="NeuralOperator1D")
class NeuralOperator1D(BaseNeuralOperator):
    """
    NeuralOperator1D, cf. [Zongyi etl al.](https://doi.org/10.48550/arXiv.2010.08895)

    Parameters
    ----------
    filters : int | list | tuple
        Number of filters for the sequential FNO layers.
        If this is a list/tuple, the length of this argument determines the number of FNO layers.
    modes : int | list | tuple
        Number of Fourier modes for FNO layers.
        An `int` results in global `modes`, a `list` allows to define the `modes` per FNO layer.
        For `rank>1`, the `modes` can be defined in terms of tuples, 
        where each entry determines the modes in the respective direction,
        e.g., `modes=(8, 4)` will result in 8 modes in y, and 4 modes in x-direction.
    input_projection_dimension : int, optional
        Projection dimension for the input layer.
        If `None`, there is no projection layer.
        Defaults to `None`.
    output_projection_dimension : int, optional
        Projection dimension for the output layer.
        If `None`, there is no projection layer.
        Defaults to `None`.
    data_format : str, optional {`"channels_first"`, `"channels_last"`}
        Data format for the convolution operations.
        Defaults to `None`.
    merge_layer : str | keras.layers.Layer, optional {`"concatenate"`, `"average"`, `"maximum"`, `"minimum"`, `"add"`, `"subtract"`, `"multiply"`, `"dot"`}
        Merge operation in FNO layers to combine the result from the spectral convolution with the result from the bypass convolution.
        Defaults to `"add"`.
    activation : str | keras.activations.Activation | keras.layers.Layer, optional
        Global activation function.
        Can be either a `str`, a `keras.activations.Activation`, or a `keras.layers.Layer`.
        Defaults to `"gelu"`.
    use_bias : bool, optional
        If `True`, all layers use a bias.
        Defaults to `True`.
    kernel_initializer : str | keras.initializers.Initializer, optional
        Kernel initializer.
        Defaults to `"he_normal"`.
    bias_initializer : str | keras.initializers.Initializer, optional
        Bias initializer.
        Defaults to `"zeros"`.
    kernel_regularizer : str | keras.regularizers.Regularizer, optional
        Kernel regularizer.
        Defaults to `None`.
    bias_regularizer : str | keras.regularizers.Regularizer, optional
        Bias regularizer.
        Defaults to `None`.
    kernel_constraint : str | keras.constraints.Constraint, optional
        Kernel constraint.
        Defaults to `None`.
    bias_constraint : str | keras.constraints.Constraint, optional
        Bias constraint.
        Defaults to `None`.
    name : str, optional
        Name of the model.
        If `None`, `name` is automatically inherited from the class name `"NeuralOperator1D"`.
        Defaults to `None`.

    """

    def __init__(
        self,
        filters,
        modes,
        input_projection_dimension=None,
        output_projection_dimension=None,
        data_format=None,
        merge_layer="add",
        activation="gelu",
        use_bias=True,
        kernel_initializer=FNOInitializer(),
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        name=None,
        **kwargs
    ):
        super().__init__(
            rank=1,
            filters=filters,
            modes=modes,
            input_projection_dimension=input_projection_dimension,
            output_projection_dimension=output_projection_dimension,
            data_format=data_format,
            merge_layer=merge_layer,
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            name=name,
            **kwargs
        )


@saving.register_keras_serializable(package="Kerex.Models.NeuralOperator", name="NeuralOperator2D")
class NeuralOperator2D(BaseNeuralOperator):
    """
    NeuralOperator2D, cf. [Zongyi etl al.](https://doi.org/10.48550/arXiv.2010.08895)

    Parameters
    ----------
    filters : int | list | tuple
        Number of filters for the sequential FNO layers.
        If this is a list/tuple, the length of this argument determines the number of FNO layers.
    modes : int | list | tuple
        Number of Fourier modes for FNO layers.
        An `int` results in global `modes`, a `list` allows to define the `modes` per FNO layer.
        For `rank>1`, the `modes` can be defined in terms of tuples, 
        where each entry determines the modes in the respective direction,
        e.g., `modes=(8, 4)` will result in 8 modes in y, and 4 modes in x-direction.
    input_projection_dimension : int, optional
        Projection dimension for the input layer.
        If `None`, there is no projection layer.
        Defaults to `None`.
    output_projection_dimension : int, optional
        Projection dimension for the output layer.
        If `None`, there is no projection layer.
        Defaults to `None`.
    data_format : str, optional {`"channels_first"`, `"channels_last"`}
        Data format for the convolution operations.
        Defaults to `None`.
    merge_layer : str | keras.layers.Layer, optional {`"concatenate"`, `"average"`, `"maximum"`, `"minimum"`, `"add"`, `"subtract"`, `"multiply"`, `"dot"`}
        Merge operation in FNO layers to combine the result from the spectral convolution with the result from the bypass convolution.
        Defaults to `"add"`.
    activation : str | keras.activations.Activation | keras.layers.Layer, optional
        Global activation function.
        Can be either a `str`, a `keras.activations.Activation`, or a `keras.layers.Layer`.
        Defaults to `"gelu"`.
    use_bias : bool, optional
        If `True`, all layers use a bias.
        Defaults to `True`.
    kernel_initializer : str | keras.initializers.Initializer, optional
        Kernel initializer.
        Defaults to `"he_normal"`.
    bias_initializer : str | keras.initializers.Initializer, optional
        Bias initializer.
        Defaults to `"zeros"`.
    kernel_regularizer : str | keras.regularizers.Regularizer, optional
        Kernel regularizer.
        Defaults to `None`.
    bias_regularizer : str | keras.regularizers.Regularizer, optional
        Bias regularizer.
        Defaults to `None`.
    kernel_constraint : str | keras.constraints.Constraint, optional
        Kernel constraint.
        Defaults to `None`.
    bias_constraint : str | keras.constraints.Constraint, optional
        Bias constraint.
        Defaults to `None`.
    name : str, optional
        Name of the model.
        If `None`, `name` is automatically inherited from the class name `"NeuralOperator2D"`.
        Defaults to `None`.

    """

    def __init__(
        self,
        filters,
        modes,
        input_projection_dimension=None,
        output_projection_dimension=None,
        data_format=None,
        merge_layer="add",
        activation="gelu",
        use_bias=True,
        kernel_initializer=FNOInitializer(),
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        name=None,
        **kwargs
    ):
        super().__init__(
            rank=2,
            filters=filters,
            modes=modes,
            input_projection_dimension=input_projection_dimension,
            output_projection_dimension=output_projection_dimension,
            data_format=data_format,
            merge_layer=merge_layer,
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            name=name,
            **kwargs
        )


@saving.register_keras_serializable(package="Kerex.Models.NeuralOperator", name="NeuralOperator3D")
class NeuralOperator3D(BaseNeuralOperator):
    """
    NeuralOperator3D, cf. [Zongyi etl al.](https://doi.org/10.48550/arXiv.2010.08895)

    Parameters
    ----------
    filters : int | list | tuple
        Number of filters for the sequential FNO layers.
        If this is a list/tuple, the length of this argument determines the number of FNO layers.
    modes : int | list | tuple
        Number of Fourier modes for FNO layers.
        An `int` results in global `modes`, a `list` allows to define the `modes` per FNO layer.
        For `rank>1`, the `modes` can be defined in terms of tuples, 
        where each entry determines the modes in the respective direction,
        e.g., `modes=(8, 4)` will result in 8 modes in y, and 4 modes in x-direction.
    input_projection_dimension : int, optional
        Projection dimension for the input layer.
        If `None`, there is no projection layer.
        Defaults to `None`.
    output_projection_dimension : int, optional
        Projection dimension for the output layer.
        If `None`, there is no projection layer.
        Defaults to `None`.
    data_format : str, optional {`"channels_first"`, `"channels_last"`}
        Data format for the convolution operations.
        Defaults to `None`.
    merge_layer : str | keras.layers.Layer, optional {`"concatenate"`, `"average"`, `"maximum"`, `"minimum"`, `"add"`, `"subtract"`, `"multiply"`, `"dot"`}
        Merge operation in FNO layers to combine the result from the spectral convolution with the result from the bypass convolution.
        Defaults to `"add"`.
    activation : str | keras.activations.Activation | keras.layers.Layer, optional
        Global activation function.
        Can be either a `str`, a `keras.activations.Activation`, or a `keras.layers.Layer`.
        Defaults to `"gelu"`.
    use_bias : bool, optional
        If `True`, all layers use a bias.
        Defaults to `True`.
    kernel_initializer : str | keras.initializers.Initializer, optional
        Kernel initializer.
        Defaults to `"he_normal"`.
    bias_initializer : str | keras.initializers.Initializer, optional
        Bias initializer.
        Defaults to `"zeros"`.
    kernel_regularizer : str | keras.regularizers.Regularizer, optional
        Kernel regularizer.
        Defaults to `None`.
    bias_regularizer : str | keras.regularizers.Regularizer, optional
        Bias regularizer.
        Defaults to `None`.
    kernel_constraint : str | keras.constraints.Constraint, optional
        Kernel constraint.
        Defaults to `None`.
    bias_constraint : str | keras.constraints.Constraint, optional
        Bias constraint.
        Defaults to `None`.
    name : str, optional
        Name of the model.
        If `None`, `name` is automatically inherited from the class name `"NeuralOperator3D"`.
        Defaults to `None`.

    """

    def __init__(
        self,
        filters,
        modes,
        input_projection_dimension=None,
        output_projection_dimension=None,
        data_format=None,
        merge_layer="add",
        activation="gelu",
        use_bias=True,
        kernel_initializer=FNOInitializer(),
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        name=None,
        **kwargs
    ):
        super().__init__(
            rank=3,
            filters=filters,
            modes=modes,
            input_projection_dimension=input_projection_dimension,
            output_projection_dimension=output_projection_dimension,
            data_format=data_format,
            merge_layer=merge_layer,
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            name=name,
            **kwargs
        )


from keras import saving
from keras import initializers
from .base_spectral_conv import BaseSpectralConv



@saving.register_keras_serializable(package="Kerex.Layers.FNO", name="SpectralConv1D")
class SpectralConv1D(BaseSpectralConv):
    """
    SpectralConv1D layer (base of FNO, cf. https://arxiv.org/abs/2010.08895)

    This layer
    (1) transforms the layer into Fourier space via discrete Fourier transform for real-valued data (rfft),
    (2) truncates the Fourier modes to the `modes` lowest modes,
    (3) applies the weights (separated into real- and imaginary weights),
    (4) pads the truncated signal in Fourier space back to its initial shape,
    (5) applies inverse discrete Fourier transform for real-valued data (irfft), and
    (6) applies the bias in physical space (if `use_bias=True`)

    Parameters
    ----------
    filters : int
        Number of filters.
    modes : int
        Number of modes after truncation in Fourier space.
    data_format : str, optional {`"channels_first"`, `"channels_last"`}
        Format of the input data.
        Defaults to `None`.
    use_bias : bool, optional
        If `True`, a bias term is added to the physical space after the spectral convolution.
        Defaults to `True`.
    kernel_initializer : str | keras.initializer.Initializer | tuple, optional
        Initializer for real- and imaginary weights.
        By default, the real weights are initialized using `"glorot_normal"`,
        and the imaginary weights are initialized using `"random"` with a low standard deviation,
        which is effectively white Gaussian noise.
        Defaults to `("glorot_normal", initializers.RandomNormal(stddev=1e-3))`.
    bias_initializer : str | keras.initializer.Initializer, optional
        Initializer for the bias.
        Defaults to `"zeros"`.
    kernel_regularizer : str | keras.regularizers.Regularizer, optional
        Kernel regularizer.
        Defaults to `None`.
    bias_regularizer : str | keras.regularizers.Regularizer, optional
        Bias regularizer.
        Defaults to `None`.
    kernel_constraint : str | keras.constraints.Constraint, optional
        Kernel constraint.
        Defaults to `None`.
    bias_constraint : str | keras.constraints.Constraint, optional
        Bias constraint.
        Defaults to `None`.
    name : str, optional
        Name of the model.
        If `None`, `name` is automatically inherited from the class name `"SpectralConv1D"`.
        Defaults to `None`.

    Notes
    -----
    For implementataion simplicity, the Fourier operations are always performed in `"channels_first"` data format.
    The layer therefore applies a transpose operation if `data_format="channels_last"`.
    
    """

    def __init__(
        self,
        filters,
        modes,
        data_format=None,
        use_bias=True,
        kernel_initializer=("glorot_normal", initializers.RandomNormal(stddev=1e-3)),
        bias_initializer="zeros",
        kernel_constraint=None,
        bias_constraint=None,
        kernel_regularizer=None,
        bias_regularizer=None,
        name=None,
        **kwargs
    ):
        super().__init__(
            rank=1,
            filters=filters,
            modes=modes,
            data_format=data_format,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            name=name,
            **kwargs
        )


@saving.register_keras_serializable(package="Kerex.Layers.FNO", name="SpectralConv2D")
class SpectralConv2D(BaseSpectralConv):
    """
    SpectralConv2D layer (base of FNO, cf. https://arxiv.org/abs/2010.08895)

    This layer
    (1) transforms the layer into Fourier space via discrete Fourier transform for real-valued data (rfft),
    (2) truncates the Fourier modes to the `modes` lowest modes,
    (3) applies the weights (separated into real- and imaginary weights),
    (4) pads the truncated signal in Fourier space back to its initial shape,
    (5) applies inverse discrete Fourier transform for real-valued data (irfft), and
    (6) applies the bias in physical space (if `use_bias=True`)

    Parameters
    ----------
    filters : int
        Number of filters.
    modes : int
        Number of modes after truncation in Fourier space.
    data_format : str, optional {`"channels_first"`, `"channels_last"`}
        Format of the input data.
        Defaults to `None`.
    use_bias : bool, optional
        If `True`, a bias term is added to the physical space after the spectral convolution.
        Defaults to `True`.
    kernel_initializer : str | keras.initializer.Initializer | tuple, optional
        Initializer for real- and imaginary weights.
        By default, the real weights are initialized using `"glorot_normal"`,
        and the imaginary weights are initialized using `"random"` with a low standard deviation,
        which is effectively white Gaussian noise.
        Defaults to `("glorot_normal", initializers.RandomNormal(stddev=1e-3))`.
    bias_initializer : str | keras.initializer.Initializer, optional
        Initializer for the bias.
        Defaults to `"zeros"`.
    kernel_regularizer : str | keras.regularizers.Regularizer, optional
        Kernel regularizer.
        Defaults to `None`.
    bias_regularizer : str | keras.regularizers.Regularizer, optional
        Bias regularizer.
        Defaults to `None`.
    kernel_constraint : str | keras.constraints.Constraint, optional
        Kernel constraint.
        Defaults to `None`.
    bias_constraint : str | keras.constraints.Constraint, optional
        Bias constraint.
        Defaults to `None`.
    name : str, optional
        Name of the model.
        If `None`, `name` is automatically inherited from the class name `"SpectralConv2D"`.
        Defaults to `None`.

    Notes
    -----
    For implementataion simplicity, the Fourier operations are always performed in `"channels_first"` data format.
    The layer therefore applies a transpose operation if `data_format="channels_last"`.
    
    """

    def __init__(
        self,
        filters,
        modes,
        data_format=None,
        use_bias=True,
        kernel_initializer=("glorot_normal", initializers.RandomNormal(stddev=1e-3)),
        bias_initializer="zeros",
        kernel_constraint=None,
        bias_constraint=None,
        kernel_regularizer=None,
        bias_regularizer=None,
        name=None,
        **kwargs
    ):
        super().__init__(
            rank=2,
            filters=filters,
            modes=modes,
            data_format=data_format,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            name=name,
            **kwargs
        )


@saving.register_keras_serializable(package="Kerex.Layers.FNO", name="SpectralConv3D")
class SpectralConv3D(BaseSpectralConv):
    """
    SpectralConv3D layer (base of FNO, cf. https://arxiv.org/abs/2010.08895)

    This layer
    (1) transforms the layer into Fourier space via discrete Fourier transform for real-valued data (rfft),
    (2) truncates the Fourier modes to the `modes` lowest modes,
    (3) applies the weights (separated into real- and imaginary weights),
    (4) pads the truncated signal in Fourier space back to its initial shape,
    (5) applies inverse discrete Fourier transform for real-valued data (irfft), and
    (6) applies the bias in physical space (if `use_bias=True`)

    Parameters
    ----------
    filters : int
        Number of filters.
    modes : int
        Number of modes after truncation in Fourier space.
    data_format : str, optional {`"channels_first"`, `"channels_last"`}
        Format of the input data.
        Defaults to `None`.
    use_bias : bool, optional
        If `True`, a bias term is added to the physical space after the spectral convolution.
        Defaults to `True`.
    kernel_initializer : str | keras.initializer.Initializer | tuple, optional
        Initializer for real- and imaginary weights.
        By default, the real weights are initialized using `"glorot_normal"`,
        and the imaginary weights are initialized using `"random"` with a low standard deviation,
        which is effectively white Gaussian noise.
        Defaults to `("glorot_normal", initializers.RandomNormal(stddev=1e-3))`.
    bias_initializer : str | keras.initializer.Initializer, optional
        Initializer for the bias.
        Defaults to `"zeros"`.
    kernel_regularizer : str | keras.regularizers.Regularizer, optional
        Kernel regularizer.
        Defaults to `None`.
    bias_regularizer : str | keras.regularizers.Regularizer, optional
        Bias regularizer.
        Defaults to `None`.
    kernel_constraint : str | keras.constraints.Constraint, optional
        Kernel constraint.
        Defaults to `None`.
    bias_constraint : str | keras.constraints.Constraint, optional
        Bias constraint.
        Defaults to `None`.
    name : str, optional
        Name of the model.
        If `None`, `name` is automatically inherited from the class name `"SpectralConv3D"`.
        Defaults to `None`.

    Notes
    -----
    For implementataion simplicity, the Fourier operations are always performed in `"channels_first"` data format.
    The layer therefore applies a transpose operation if `data_format="channels_last"`.
    
    """

    def __init__(
        self,
        filters,
        modes,
        data_format=None,
        use_bias=True,
        kernel_initializer=("glorot_normal", initializers.RandomNormal(stddev=1e-3)),
        bias_initializer="zeros",
        kernel_constraint=None,
        bias_constraint=None,
        kernel_regularizer=None,
        bias_regularizer=None,
        name=None,
        **kwargs
    ):
        super().__init__(
            rank=3,
            filters=filters,
            modes=modes,
            data_format=data_format,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            name=name,
            **kwargs
        )
